##################################################################################
# This file is part of NNERO.
#
# Copyright (c) 2024, Gaétan Facchinetti
#
# NNERO is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. NNERO is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with NNERO. 
# If not, see <https://www.gnu.org/licenses/>.
#
##################################################################################


import random
import numpy as np
import torch
from scipy import interpolate
from os.path import abspath, exists

from typing import Self

from .cosmology import optical_depth_no_rad


LATEX_LABELS = {'hlittle' : r'$h$', 'Ln_1010_As' : r'$\ln (10^{10}A_{\rm s})$', 'F_STAR10' : r'$\log_{10} f_{\star, 10}$',
                  'ALPHA_STAR' : r'$\alpha_\star$', 't_STAR' : r'$t_\star$', 'F_ESC10' : r'$\log_{10} f_{\rm esc, 10} $', 
                  'ALPHA_ESC' : r'$\alpha_{\rm esc}$', 'M_TURN' : r'$\log_{10} \left[\frac{M_{\rm turn}}{M_\odot}\right]$', 'Omdmh2' : r'$\omega_{\rm dm}$', 
                  'Ombh2' : r'$\omega_{\rm b}$', 'POWER_INDEX' : r'$n_{\rm s}$', 'M_WDM' : r'$m_{\rm WDM}~{\rm [keV]}$', 
                  'INVERSE_M_WDM' : r'$\mu_{\rm WDM}$', 'FRAC_WDM' : r'$f_{\rm WDM}$', 'NEUTRINO_MASS_1' : r'$m_{\nu_1}$',
                  'LOG10_PMF_SB' : r'$\log_{10} s_{\rm B}$', 'PMF_NB' : r'$n_{\rm B}$', 'tau_reio' : r'$\tau$', 
                  'L_X' : r'$\log_{10} L_X$', 'NU_X_THRESH' : r'$E_0$', 'sum_mnu' : r'$\sum {m_\nu}~{\rm [eV]}$', 
                  'LOG10_PMF_SIGMA_B0' : r'$\log_{10}{\sigma_{\rm B, 0}}$'}


MP_KEY_CORRESPONDANCE = {'log10_f_star10' : 'F_STAR10', 'alpha_star' : 'ALPHA_STAR', 't_star' : 't_STAR', 'log10_f_esc10' : 'F_ESC10', 
                         'alpha_esc' : 'ALPHA_ESC', 'Omch2' : 'Omdmh2', 'omega_dm' : 'Omdmh2', 'omega_b' : 'Ombh2', 'h': 'hlittle', 'ln10^{10}A_s' : 'Ln_1010_As',
                         'n_s' : 'POWER_INDEX', 'm_nu1' : 'NEUTRINO_MASS_1', 'f_wdm' : 'FRAC_WDM', 'm_wdm' : 'M_WDM', 'nu_X_thresh' : 'NU_X_THRESH',
                         'log10_pmf_sb' : 'LOG10_PMF_SB', 'pmf_nb' : 'PMF_NB', 'log10_m_turn' : 'M_TURN', 'log10_lum_X' : 'L_X', '1/m_wdm' : 'INVERSE_M_WDM',
                         'log10_pmf_sigma_b0' : 'LOG10_PMF_SIGMA_B0'}

#LATEX_LABELS_CLASS = {'omega_b' :  r'$\omega_{\rm b}$', 'omega_dm' : r'$\omega_{\rm dm}$', 'h' : r'$h$', 'ln10^{10}A_s' : r'$\ln 10^{10} A_{\rm s}$',
#                'n_s' : r'$n_{\rm s}$', 'm_nu1' : r'$m_{\nu 1}~{\rm [eV]}$', 'sum_mnu' : r'$\sum {m_\nu}~{\rm [eV]}$', 'log10_f_star10' : r'$\log_{10}f_{\star, 10}$',
#                  'alpha_star' : r'$\alpha_\star$', 'log10_f_esc10' : r'$\log_{10} f_{\rm esc, 10}$', 'alpha_esc' : r'$\alpha_{\rm esc}$',
#                  't_star' : r'$t_\star$', 'log10_m_turn' : r'$\log_{10} M_{\rm turn}$', 'log10_lum_X' : r'$\log_{10} L_X$', 'nu_X_thresh' : r'$E_0$',
#                  '1/m_wdm' : r'$\mu_{\rm WDM}$', 'mu_wdm' : r'$\mu_{\rm WDM}$', 'f_wdm' : r'$f_{\rm WDM}$', 'tau_reio' : r'$\tau$'}


def latex_labels(labels: list[str]) -> list[str]:
    
    res = [None] * len(labels)

    for il, label in enumerate(labels):

        if label in LATEX_LABELS.keys(): 
            res[il] = LATEX_LABELS[label]
        elif label in MP_KEY_CORRESPONDANCE.keys():  
            if MP_KEY_CORRESPONDANCE[label] in LATEX_LABELS.keys():
                res[il] = LATEX_LABELS[MP_KEY_CORRESPONDANCE[label]]
            else:
                res[il] = label
        else:
            res[il] = label
        
    return res


def preprocess_raw_data(file_path: str, *, random_seed: int = 1994, frac_test: float = 0.1, frac_valid: float = 0.1, extras: list[str] | None = None) -> None:
    """
    Preprocess a raw .npz file. 
    Creates another numpy archive that can be directly used to create a :py:class:`DataSet` object.

    Parameters
    ----------
    file_path : str
        Path to the raw data file. The raw data must be a .npz file with the following information. 
        - z (or z_glob): redshift array
        - features_run:  Sequence of drawn input parameters for which there is a value for the ionization fraction
        - features_fail: Sequence of drawn input parameters for which the simulator failed because reionization was too late
        - ...
    random_seed : int, optional
        Random seed for splitting data into a training/validation/testing subset, by default 1994.
    frac_test : float, optional
       Fraction of the total data points in the test subset, by default 0.1.
    frac_valid : float, optional
        Fraction of the total data points in the validation subset, by default 0.1
    """

    # start by setting the random seed
    random.seed(random_seed)

    with open(file_path, 'rb') as file:

        # data must have been stored in a numpy archive with the correct format
        data = np.load(file, allow_pickle=True)
        z_glob               = data.get('z_glob', None)
        features_run         = data.get('features_run', None)
        features_super_late  = data.get('features_late', None)
        cosmology_run        = data.get('cosmology_run', None)
        cosmology_super_late = data.get('cosmology_late', None)
        parameters_min_val   = data.get('parameters_min_val', None)
        parameters_max_val   = data.get('parameters_max_val', None)
        xHIIdb               = data.get('xHIIdb', None)
        parameters_name      = data.get('parameters_name', None)
        ps                   = data.get('ps', None)
        
        extras_array = None
        if extras is not None and features_run is not None:
            extras_array = np.zeros((len(extras), features_run.shape[0]))
            for iex, ex in enumerate(extras):
                extras_array[iex] = data.get(ex, np.zeros(features_run.shape[0]))

        # Check for different notations
        z_glob = z_glob if (z_glob is not None) else data.get('z', None)
        features_super_late = features_super_late if (features_super_late is not None) else data.get('features_fail', None)

    # --------------------------

    new_redshifts = [5.9]

    # add some values we may need for the data classification
    new_z = np.sort(new_redshifts)
    for _z in new_z: 
        if _z not in z_glob:
            pos = np.searchsorted(z_glob, _z)
            value = interpolate.interp1d(z_glob, xHIIdb, kind='slinear')(_z)
            z_glob = np.sort(np.append(z_glob, _z))
            xHIIdb = np.insert(xHIIdb, pos, value, axis=1)


    features  = np.vstack((features_run, features_super_late))
    cosmology = np.concatenate((cosmology_run, cosmology_super_late)) 

    # total number of features
    n_r  = features_run.shape[0]
    n_sl = features_super_late.shape[0]
    n_tot = n_r + n_sl

    # parameters for xHIIdb
    xHIIdb = np.vstack((xHIIdb, np.zeros((n_sl, xHIIdb.shape[1]))))

    if ps is not None: ps = np.concatenate((ps, np.zeros((n_sl, ps.shape[1], ps.shape[2]))))

    if extras is not None:
        extras_array = np.hstack((extras_array, np.zeros((extras_array.shape[0], n_sl))))

    # shuffling all the data between late and run
    r = random.sample(range(n_tot), n_tot)

    features  = features[r]
    cosmology = cosmology[r]
    xHIIdb    = xHIIdb[r]
    ps        = ps[r]

    if extras is not None:
        for iex, _ in enumerate(extras_array):
            extras_array[iex] = extras_array[iex, r]

    # data selection, only considering the "early time" reionizations
    pos = np.searchsorted(z_glob, 5.9) 

    # define the early time reionization quantities
    indices_early = np.where(xHIIdb[:, pos] > 0.69)[0]

    # divide the early data into train, test and validation datasets
    # to that end shuffles indices early and grab slices of the shuffled dataset
    n_early = len(indices_early)
    r_early = random.sample(range(n_early), n_early)
    r_indices_early = indices_early[r_early]

    n_early_test  = int(frac_test*n_early)
    n_early_valid = int(frac_valid*n_early)

    indices_early_test  = np.sort(r_indices_early[:n_early_test])
    indices_early_valid = np.sort(r_indices_early[n_early_test:(n_early_test+n_early_valid)])
    indices_early_train = np.sort(r_indices_early[(n_early_test+n_early_valid):])

    # devide now the entire data into train, test and validation datasets
    r_indices_tot = random.sample(range(n_tot), n_tot)

    n_tot_test  = int(frac_test*n_tot)
    n_tot_valid = int(frac_valid*n_tot)

    indices_total_test  = np.sort(r_indices_tot[:n_tot_test])
    indices_total_valid = np.sort(r_indices_tot[n_tot_test:(n_tot_test+n_tot_valid)])
    indices_total_train = np.sort(r_indices_tot[(n_tot_test+n_tot_valid):])

    # save in file with _pp extension standing for "preprocessed"
    with open(file_path[:-4] + "_pp.npz", 'wb') as file:

        # data must have been stored in a numpy archive with the correct format
        np.savez(file, 
                 redshifts = z_glob, 
                 features = features, 
                 cosmology = cosmology, 
                 xHIIdb = xHIIdb, 
                 parameters_min_val = parameters_min_val, 
                 parameters_max_val = parameters_max_val, 
                 parameters_name = parameters_name,
                 indices_early_test  = indices_early_test,
                 indices_early_valid = indices_early_valid,
                 indices_early_train = indices_early_train,
                 indices_total_test  = indices_total_test,
                 indices_total_valid = indices_total_valid,
                 indices_total_train = indices_total_train,
                 random_seed = random_seed,
                 frac_test = frac_test,
                 frac_valid = frac_valid,
                 extras_array = extras_array,
                 extras_name  = extras,
                 ps = ps)


def true_to_uniform(x: float | np.ndarray,
                    min: float | np.ndarray,
                    max: float | np.ndarray) -> float | np.ndarray:
    """
    Transforms features uniformely distributed along [a, b] into features
    uniformely distributed between [0, 1] as fed to the neural networks.

    Parameters
    ----------
    x : float | np.ndarray
        input featurs distributed uniformely on [a, b]
    min : float | np.ndarray
        minimum value `a`
    max : float | np.ndarray
        maximum value `b`

    Returns
    -------
    float | np.ndarray

    Raises
    ------
    ValueError
        min should be less than max
    """

    if not np.all(min <= max):
        raise ValueError("The minimum value is bigger than the maximum one")
    
    return (x - min) / (max - min)

def uniform_to_true(x: float | np.ndarray,
                    min: float | np.ndarray,
                    max: float | np.ndarray) -> float | np.ndarray:
    """
    Inverse transformation of `true_to_uniform`.

    Parameters
    ----------
    x : float | np.ndarray
        input featurs distributed uniformely on [0, 1]
    min : float | np.ndarray
        minimum value `a`
    max : float | np.ndarray
        maximum value `b`

    Returns
    -------
    float | np.ndarray

    Raises
    ------
    ValueError
        min should be less than max
    """
    
    if not np.all(min <= max):
        raise ValueError("The minimum value is bigger than the maximum one")
    
    return (max - min) * x + min



class DataPartition:
    """
    DataPartition class. 

    Partitioning of the data into a training set, a testing set and a validation set.

    Parameters
    ----------
    early_train : np.ndarray
        indices of the data array with an early enough reionization used for training
    early_valid : np.ndarray
        indices of the data array with an early enough reionization used for validation
    early_test : np.ndarray
        indices of the data array with an early enough reionization used for testing
    total_train : np.ndarray
        all indices of the data array used for training
    total_valid : np.ndarray
        all indices of the data array used for validation
    total_test : np.ndarray
        all indices of the data array used for testing
    """

    def __init__(self, 
                 early_train: np.ndarray,
                 early_valid: np.ndarray, 
                 early_test: np.ndarray, 
                 total_train: np.ndarray, 
                 total_valid: np.ndarray, 
                 total_test: np.ndarray) -> None:
        
        self._early_dict = {'train' : early_train, 'valid': early_valid, 'test' : early_test}
        self._total_dict = {'train' : total_train, 'valid': total_valid, 'test' : total_test}
        
        self._early = np.sort(np.concatenate((self.early_test, self.early_valid, self.early_train)))

        # by default the selection is the full set of early points
        self.set_selection(self._early)

    def __call__(self):
        _new_early = {('early_' + k): val for k, val in self._early_dict.items()}
        _new_total = {('total_' + k): val for k, val in self._total_dict.items()}
        return (_new_early | _new_total)
    
    def __eq__(self, other):

        other_dict = other()

        for key, val in self().items():
            if (val is not None) and (other_dict[key] is not None):
                if len(val) != len(other_dict[key]): 
                    return False
                if np.any(val != other_dict[key]):
                    return False
            if (val is None) and (other_dict[key] is not None):
                return False
            if  (val is not None) and (other_dict[key] is None):
                return False
            
        return True
    
    def save(self, name:str) -> None:
        """
        Save the data partition.

        Parameters
        ----------
        name : str
            name of the data partition file
        """

        with open(name + '.npz', 'wb') as file:
            np.savez(file = file, 
                     early_train = self.early_train, 
                     early_valid = self.early_valid,
                     early_test  = self.early_test,
                     total_train = self.total_train,
                     total_valid = self.total_valid,
                     total_test  = self.total_test)

    @classmethod
    def load(cls, path: str) -> Self:
        """
        Load a previously saved data partition.

        Parameters
        ----------
        path : str
            path to the data partition saved file.

        Returns
        -------
        DataPartition
        """
        
        with open(path + '.npz', 'rb') as file:
            data = np.load(file, allow_pickle=False)    
            return DataPartition(data.get('early_train'), 
                                 data.get('early_valid'), 
                                 data.get('early_test'), 
                                 data.get('total_train'), 
                                 data.get('total_valid'), 
                                 data.get('total_test'))
        

    def set_selection(self, sample: np.ndarray) -> None:

        # define a subset of the early partition
        # this "selection" is what the regressor is trained on

        self._selection_train = np.intersect1d(sample, self.early_train)
        self._selection_valid = np.intersect1d(sample, self.early_valid)
        self._selection_test  = np.intersect1d(sample, self.early_test)

        self._selection = np.intersect1d(sample, self.early)
    

    @property
    def early_train(self):
        return self._early_dict['train']
    
    @property
    def early_valid(self):
        return self._early_dict['valid']
    
    @property
    def early_test(self):
        return self._early_dict['test']
    
    @property
    def total_train(self):
        return self._total_dict['train']
    
    @property
    def total_valid(self):
        return self._total_dict['valid']
    
    @property
    def total_test(self):
        return self._total_dict['test']
    
    @property
    def selection(self):
        return self._selection
    
    @property
    def selection_train(self):
        return self._selection_train
    
    @property
    def selection_valid(self):
        return self._selection_valid
    
    @property
    def selection_test(self):
        return self._selection_test
    
    @property
    def early(self):
        return self._early
    


class MetaData:
    """
    MetaData class
    
    Metadata that is saved with the neural network for predictions.

    Parameters
    ----------
    z : np.ndarray
        array of redshifts
    parameters_name : list | np.ndarray
        name of the parameters (input features)
    parameters_min_val : np.ndarray
        minimum value of the parameters (input features)
    parameters_max_val : np.ndarray
        maximum value of the parameters (input features)
    """

    def __init__(self,
                 z: np.ndarray,
                 parameters_name: list | np.ndarray,
                 parameters_min_val: np.ndarray,
                 parameters_max_val: np.ndarray) -> None:
       
        self._z                  = z
        self._parameters_name    = np.array(parameters_name)
        self._parameters_min_val = parameters_min_val
        self._parameters_max_val = parameters_max_val

        # derived quantities
        self._pos_omega_b  = np.where(self.parameters_name == 'Ombh2')[0][0]
        self._pos_omega_dm = np.where(self.parameters_name == 'Omdmh2')[0][0]
        self._pos_hlittle  = np.where(self.parameters_name == 'hlittle')[0][0]

        self._min_omega_b = self._parameters_min_val[self._pos_omega_b]
        self._min_omega_c = self._parameters_min_val[self._pos_omega_dm]
        self._min_hlittle = self._parameters_min_val[self._pos_hlittle]

        self._max_omega_b = self._parameters_max_val[self._pos_omega_b]
        self._max_omega_c = self._parameters_max_val[self._pos_omega_dm]
        self._max_hlittle = self._parameters_max_val[self._pos_hlittle]

        # principal component analysis quantities

        self._pca_mean_values  = np.empty(0)
        self._pca_eigenvalues  = np.empty(0)
        self._pca_eigenvectors = np.empty((0, 0))
        self._pca_n_eigenvectors = 0

        self._torch_pca_mean_values  = torch.tensor(self._pca_mean_values,  dtype=torch.float32)
        self._torch_pca_eigenvectors = torch.tensor(self._pca_eigenvectors, dtype=torch.float32)
        
    

    def __call__(self):
        return {'z' : self._z, 
                'parameters_name'    : self._parameters_name, 
                'parameters_min_val' : self._parameters_min_val,
                'parameters_max_val' : self._parameters_max_val,
                'pca_eigenvalues'    : self._pca_eigenvalues,
                'pca_eigenvectors'   : self._pca_eigenvectors,
                'pca_mean_values'    : self._pca_mean_values}

    def __eq__(self, other):
        
        other_dict = other()

        for key, val in self().items():
            if (val is not None) and (other_dict[key] is not None):
                if len(val) != len(other_dict[key]): 
                    return False
                if np.any(val != other_dict[key]):
                    return False
            if (val is None) and (other_dict[key] is not None):
                return False
            if  (val is not None) and (other_dict[key] is None):
                return False
        
        if self.pca_n_eigenvectors != other.pca_n_eigenvectors:
            return False
            
        return True
    
    def save(self, name: str) -> None:
        """
        Save the metadata.

        Parameters
        ----------
        name : str
            name of the metadata file
        """

        with open(name + '.npz', 'wb') as file:
            np.savez(file = file, 
                     z = self.z, 
                     parameters_name = self.parameters_name,
                     parameters_min_val = self.parameters_min_val,
                     parameters_max_val = self.parameters_max_val,
                     pca_mean_values    = self.pca_mean_values,
                     pca_eigenvalues    = self.pca_eigenvalues,
                     pca_eigenvectors   = self.pca_eigenvectors,
                     pca_n_eigenvectors = self.pca_n_eigenvectors)

    @classmethod
    def load(cls, path: str) -> Self:
        """
        Load a previously saved metadata file.

        Parameters
        ----------
        path : str
            path to the metadata saved file.

        Returns
        -------
        MetaData
        """
        
        with open(path + '.npz', 'rb') as file:
            data = np.load(file, allow_pickle=True)    
            metadata = MetaData(data.get('z'), 
                            data.get('parameters_name'), 
                            data.get('parameters_min_val'), 
                            data.get('parameters_max_val'))
            
            # get the pca decomposition
            metadata._pca_eigenvalues    = data.get('pca_eigenvalues',    np.empty(0))
            metadata._pca_eigenvectors   = data.get('pca_eigenvectors',   np.empty(0))
            metadata._pca_mean_values    = data.get('pca_mean_values',    np.empty(0))
            metadata._pca_n_eigenvectors = data.get('pca_n_eigenvectors', 0)

            metadata._torch_pca_mean_values  = torch.tensor(metadata.pca_mean_values,  dtype=torch.float32)
            metadata._torch_pca_eigenvectors = torch.tensor(metadata.pca_eigenvectors, dtype=torch.float32)

            return metadata

    @property
    def z(self):
        return self._z
    
    @property
    def parameters_name(self):
        return self._parameters_name
    
    @property
    def parameters_min_val(self):
        return self._parameters_min_val
    
    @property
    def parameters_max_val(self):
        return self._parameters_max_val
    
    @property
    def pos_omega_b(self):
        return self._pos_omega_b
    
    @property
    def pos_omega_dm(self):
        return self._pos_omega_dm
    
    @property
    def pos_hlittle(self):
        return self._pos_hlittle
    
    @property
    def min_omega_b(self):
        return self._min_omega_b
    
    @property
    def min_omega_c(self):
        return self._min_omega_c
    
    @property
    def min_hlittle(self):
        return self._min_hlittle
    
    @property
    def max_omega_b(self):
        return self._max_omega_b
    
    @property
    def max_omega_c(self):
        return self._max_omega_c
    
    @property
    def max_hlittle(self):
        return self._max_hlittle
    
    @property
    def pca_mean_values(self):
        return self._pca_mean_values
    
    @property
    def pca_eigenvalues(self):
        return self._pca_eigenvalues
    
    @property
    def pca_eigenvectors(self):
        return self._pca_eigenvectors
    
    @property
    def pca_n_eigenvectors(self):
        return self._pca_n_eigenvectors
    
    @property
    def torch_pca_eigenvectors(self):
        return self._torch_pca_eigenvectors
    
    @property
    def torch_pca_mean_values(self):
        return self._torch_pca_mean_values
    




class DataSet:
    """
    DataSet class

    Compile the data necessary for training.

    Parameters
    ----------
    file_path: str
        path to the file that contains the raw data
    z: np.ndarray
        array of the redshits of interpolation of the nn
    use_PCA: bool, optional
        prepare the data to perform the regression in the principal component basis, default is True
    precision_PCA: float, optional
        if use_PCA is `True`, select the number of useful eigenvectors from this coefficient 
        -- only the eigenvectors with eigenvalues larger than precision_PCA * the largest eigenvalue
        are considered as useful
    frac_test: float, optional 
        fraction of test data out of the total sample, default is 0.1
    frac_valid: float, optional
        fraction of validation data out of the total sample, default is 0.1
    seed_split: int, optional
        random seed for data partitioning, default is 1994
    """

    def __init__(self, 
                 file_path : str, 
                 z : np.ndarray | None = None, 
                 *, 
                 frac_test: float  = 0.1, 
                 frac_valid: float = 0.1,
                 seed_split: int   = 1994,
                 extras: list[str] | None = None,
                 constr_tau: float | None = None) -> None:

        # --------------------------------
        # initialisation from input values 

        # directory was the data is stored
        self._file_path = abspath(file_path)

        # define a default redshift array on which to make the predictions
        # define the labels of the regressor
        if z is None:
            _z = np.array([4, 4.25, 4.5, 4.75, 5, 5.25, 5.5, 5.75, 
                                        5.9, 6.25, 6.5, 6.75, 7, 7.25, 7.5, 7.75, 
                                        8, 8.25, 8.5, 8.75, 9, 9.5, 10, 10.5, 11, 
                                        11.5, 12, 12.5, 13, 13.5, 14, 14.5, 15, 
                                        15.5, 16, 17, 18, 19, 20, 21, 22, 23, 24, 
                                        25, 26, 27, 29, 31, 33, 35])
        else:
            _z = z


        # -----------------------------
        # prepare and read the datafile

        # if raw data has not yet been preprocessed
        if not exists(file_path[:-4]+ "_pp.npz"):
            preprocess_raw_data(file_path, random_seed=seed_split, frac_test = frac_test, frac_valid = frac_valid, extras = extras)
        else:
            with open(file_path[:-4]+ "_pp.npz", 'rb') as file:
                data = np.load(file, allow_pickle=True)
                
                # if we do not have the same seed or fraction of valid and test samples we preprocess the data again
                if frac_test != data.get('frac_test', None) or frac_valid != data.get('frac_valid', None) or seed_split != data.get('random_seed', None):
                    preprocess_raw_data(file_path, random_seed=seed_split, frac_test = frac_test, frac_valid = frac_valid, extras = extras)


        self._extras_array = None
        self._extras_name  = None

        with open(file_path[:-4]+ "_pp.npz", 'rb') as file:

            data = np.load(file, allow_pickle=True)
            
            self._redshifts = data.get('redshifts', None)
            self._features  = data.get('features',  None)
            self._cosmology = data.get('cosmology', None)
            self._xHIIdb    = data.get('xHIIdb',    None)
            self._ps        = data.get('ps', None)

            # possibility to add extra values for each run
            if isinstance(data.get('extras_array', None), np.ndarray):
                self._extras_array = data.get('extras_array')
                self._extras_name  = data.get('extras_name', None)

            # define a metadata object
            self._metadata = MetaData(_z, 
                                      data.get('parameters_name',    None),
                                      data.get('parameters_min_val', None),
                                      data.get('parameters_max_val', None))

            # define a partition object
            self._partition = DataPartition(data.get('indices_early_train', None),
                                            data.get('indices_early_valid', None),
                                            data.get('indices_early_test',  None),
                                            data.get('indices_total_train', None),
                                            data.get('indices_total_valid', None),
                                            data.get('indices_total_test',  None))        

        n_tot = len(self._features)

        # evaluate the optical depth to reionization for all runs
        # this is done with an optimised function for the evaluation of tau with numpy arrays
        # assume a late time universe with no radiation (very good approximation)
        self._tau = optical_depth_no_rad(self._redshifts, self._xHIIdb, 
                                                     self._features[:, self.metadata.pos_omega_b], 
                                                     self._features[:, self.metadata.pos_omega_dm],
                                                     self._features[:, self.metadata.pos_hlittle])
        
        self._x_array = true_to_uniform(self._features, self.metadata.parameters_min_val, self.metadata.parameters_max_val)
        

        # add a possible cut on tau to the classifier
        # so that we do not need to learn very excluded scenarios
        # with the regressor
        if constr_tau is not None:
            indices_low_tau  = np.where(self._tau < constr_tau)[0]
            indices_low_xHII = np.where(self._xHIIdb[:, -1] < 0.99)[0]
            indices_good     = np.intersect1d(indices_low_tau, indices_low_xHII)
            self.partition.set_selection(indices_good)

        # set 0 to the late reionization and 1 to that are early enough
        self._y_classifier = np.zeros(len(self._features))
        self._y_classifier[self.partition.selection] = 1.0
           
        self._y_regressor = np.zeros((n_tot, len(self.metadata.z) + 1))
        for i in range(n_tot):
            self._y_regressor[i, -1] = self._tau[i]
            if i in self.partition.selection:
                self._y_regressor[i, :-1] = interpolate.interp1d(self._redshifts, self._xHIIdb[i, :])(self.metadata.z)
            
        # convert to float32 objects
        self._x_array      = np.array(self._x_array, np.float32)
        self._y_classifier = np.array(self._y_classifier, np.float32)
        self._y_regressor  = np.array(self._y_regressor, np.float32)

        if self._extras_array is not None:
            self._extras_array = np.array(self._extras_array, np.float32)
        # --------------------------


    def init_principal_components(self, pca_precision:float = 1e-3) -> int:
        """
        Initialise the principal component analysis decomposition

        Parameters
        ----------
        pca_precision : float, optional
            precision for the principal analysis reconstruction, by default 1e-3

        Returns
        -------
        int
            number of necessary eigenvectors to reach the desired precision
        """

        # array on which we perform the principal component analysis
        arr = interpolate.interp1d(self._redshifts, np.log10(self._xHIIdb[self.partition.selection_train, :]))(self.metadata.z)
        
        # mean of the functions
        pca_mean_values = np.mean(arr, axis=0)

        # shift the function to a centered distribution
        arr_centered = arr - pca_mean_values

        # self-covariance matrix
        cov = np.cov(arr_centered, rowvar=False)

        # eigenvalues of the covariance operator
        eigenvalues, eigenvectors = np.linalg.eigh(cov)

        # reorganise the eigenvector from decreasing eigenvalues
        # and transpose the eigenvector matrix
        # now eigenvector[i, :] is the i-th eigenvector
        idx = np.argsort(eigenvalues)[::-1]
        pca_eigenvalues  = eigenvalues[idx]
        pca_eigenvectors = eigenvectors[:, idx].T

        # define the number of vectors from the precision on the eigenvalues
        pca_n_eigenvectors = np.where(np.sqrt(pca_eigenvalues/pca_eigenvalues[0]) < pca_precision)[0][0]

        # update the metadata accordingly
        self.metadata._pca_mean_values    = pca_mean_values
        self.metadata._pca_eigenvalues    = pca_eigenvalues
        self.metadata._pca_eigenvectors   = pca_eigenvectors
        self.metadata._pca_n_eigenvectors = pca_n_eigenvectors

        # define de torch version of the eigenvectors and mean values
        self.metadata._torch_pca_mean_values  = torch.tensor(self.metadata.pca_mean_values,  dtype=torch.float32)
        self.metadata._torch_pca_eigenvectors = torch.tensor(self.metadata.pca_eigenvectors, dtype=torch.float32)

        return pca_n_eigenvectors
    

    @property
    def z(self):
        return self._z
    
    @property
    def x_array(self):
        return self._x_array
    
    @property
    def y_classifier(self):
        return self._y_classifier
    
    @property
    def y_regressor(self):
        return self._y_regressor
    
    @property
    def metadata(self):
        return self._metadata
    
    @property
    def partition(self):
        return self._partition

    @property
    def tau(self):
        return self._tau
    
    @property
    def extras_array(self):
        return self._extras_array
    
    @property
    def extras_name(self):
        return self._extras_name
    
    @property
    def ps(self):
        return self._ps
    



class TorchDataset(torch.utils.data.Dataset):
    """
    Wrapper of torch Dataset.

    Parameters
    ----------
    x_data : np.ndarray
        input features
    y_data : np.ndarray
        output labels
    """
    
    def __init__(self, x_data: np.ndarray, y_data: np.ndarray) -> None:
        self.x_data = x_data
        self.y_data = y_data
    
    def __len__(self):
        return len(self.x_data)
    
    def __getitem__(self, idx: int) -> tuple:
        x = self.x_data[idx]
        y = self.y_data[idx]
        return x, y







##################################################################################
# This file is part of NNERO.
#
# Copyright (c) 2024, Gaétan Facchinetti
#
# NNERO is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. NNERO is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with NNERO. 
# If not, see <https://www.gnu.org/licenses/>.
#
##################################################################################

import numpy as np
import warnings

import os, pickle
import pkg_resources

from abc import ABC, abstractmethod
from typing import Callable

from scipy       import special, interpolate
from .data       import uniform_to_true
from .predictor  import predict_tau_from_xHII_numpy, predict_xHII_numpy, DEFAULT_VALUES
from .classifier import Classifier
from .regressor  import Regressor

from .astrophysics import m_halo, phi_uv, dn_dm
from .cosmology import h_factor_no_rad, ShortPowerSpectrumRange
from .constants import CONVERSIONS, CST_MSOL_MPC

CLASSY_IMPORTED = False

try:
    import classy
    CLASSY_IMPORTED = True
except:
    pass


LIKELIHOOD_DATA_PATH = pkg_resources.resource_filename('nnero', 'lkl_data/')


######################
## Define likelihood at fixed cosmology

class Likelihood(ABC):

    def __init__(self, parameters: list[str]) -> None:

        self._parameters = parameters
        self._index = {key: i for i, key in enumerate(parameters)} 


    def get_x(self,  theta: np.ndarray, xi: np.ndarray) -> np.ndarray:
        
        if len(theta.shape) == 1:
            theta = theta[None, :]

        if len(xi.shape) == 1:
            xi = np.tile(xi, (theta.shape[0], 1))

        return np.hstack((theta, xi))
    

    def get_x_dict(self, x: np.ndarray) -> dict:
        return {param:x[:,self._index[param]] for param in self._parameters}

    def get_x_dicts(self, x:np.ndarray) -> list[dict]:
        return [ {param:x[i,self._index[param]] for param in self._parameters} for i in range(x.shape[0])]
            
   
    def loglkl(self, theta: np.ndarray, xi: np.ndarray, **kwargs) -> np.ndarray:

        x = self.get_x(theta, xi)

        if x.shape[-1] != len(self.parameters):
            raise ValueError('theta and xi should make an array that is the size of the parameter vector' )
        
        return self._loglkl(x, **kwargs)


    @abstractmethod
    def _loglkl(self, x, **kwargs):
        pass

    @property
    def parameters(self):
        return self._parameters
    
    @property
    def index(self):
        return self._index
    
    @property
    def vectorizable(self):
        return self._vectorizable


class UVLFLikelihood(Likelihood):
    """
    Likelihood for the UV luminosity functions.

    Parameters
    ----------
    k : np.ndarray
        Array of modes on which the matter power spectrum is given (in 1/Mpc).
    pk : np.ndarray
        Matter power spectrum.
    """

    def __init__(self, 
                 parameters: list[str],  
                 *,
                 parameters_xi: np.ndarray | None = None,
                 xi: np.ndarray | None = None,
                 k: np.ndarray | None = None, 
                 pk: np.ndarray | None = None,
                 precompute: bool = False,
                 **kwargs) -> None:
    
    
        self._k  = k
        self._pk = pk

        if precompute:

            if (self._k is None) and (self._pk is None):

                if CLASSY_IMPORTED is True:

                    h = 0.7
                    omega_m = 0.3*(h**2)
                    cosmo = classy.Class()
                    
                    params_cosmo = {'output' : 'mPk', 'P_k_max_h/Mpc' : 650.0, 'h' : 0.7, 'omega_m' : omega_m}

                    if parameters_xi is not None and 'Ln1010As' in parameters_xi:
                        params_cosmo['ln10^{10}A_s'] = xi[parameters_xi.index('Ln1010As')]
                    
                    if parameters_xi is not None and 'Ombh2' in parameters_xi:
                        params_cosmo['omega_b'] = xi[parameters_xi.index('Ombh2')]

                    if parameters_xi is not None and 'POWER_INDEX' in parameters_xi:
                        params_cosmo['n_s'] = xi[parameters_xi.index('POWER_INDEX')]

                    if parameters_xi is None:
                        print("Attention: matter power spectrum evaluated for a default cosmology.")

                    cosmo.set(params_cosmo)
                    cosmo.compute()

                    self._k     = np.logspace(-2.5, np.log10(cosmo.pars['P_k_max_h/Mpc'] * h), 50000)
                    self._pk    = np.array([cosmo.pk_lin(_k, 0) for _k in self._k]) 

                else: 
                    raise ValueError("Need to import CLASS to pecompute the matter power spectrum if not given as input.")
        
        self.sheth_a = kwargs.get('sheth_a', 0.322)
        self.sheth_q = kwargs.get('sheth_q', 1.0)
        self.sheth_p = kwargs.get('sheth_p', 0.3)
        self.c       = kwargs.get('c', 2.5)
        self.window  = kwargs.get('window', 'sharpk')

        elements_to_check = ['ALPHA_STAR', 't_STAR', 'F_STAR10', 'M_TURN']
        if not np.all([element in parameters for element in elements_to_check]):
            raise ValueError('In order to use UVLFLikelihood, need to pass alpha_star, t_star, log10_f_star10 and log10_m_turn as inpur parameters')

        with open(os.path.join(LIKELIHOOD_DATA_PATH, "UVData.pkl"), 'rb') as file:
        
            data_uv = pickle.load(file)

            self.z_uv_exp = data_uv['z_uv']
            self.m_uv_exp = data_uv['m_uv']

            self.phi_uv_exp = data_uv['p_uv']

            self.sigma_phi_uv_down = data_uv['sigma_p_uv_down']
            self.sigma_phi_uv_up   = data_uv['sigma_p_uv_up']

        self._z_array = []
        self._z_index_from_table = [[] for j in [0, 1, 2, 3]]

        self._precompute = precompute
        self._masses = np.empty(0)
        self._dndmh  = np.empty(0)

        # if we can precompute the halo mass function 
        # because the matter power spectrum does not vary
        # then we do it here
        if self._precompute:

            for j in [0, 1, 2, 3]:

                # loop on the redshift bins
                for z in self.z_uv_exp[j]:

                    if z not in self._z_array:
                        self._z_array.append(z)
                    
                    self._z_index_from_table[j].append(self._z_array.index(z))


            self._z_array = np.array(self._z_array)

            rhom0  = (0.3*(0.7**2)) * CST_MSOL_MPC.rho_c_over_h2        
            m_min  = 1.1 * 4.0*np.pi/3.0 * rhom0 / (self.k[-1] / self.c)**3
            m_max  = 4.0*np.pi/3.0 * rhom0 / (self.k[0]/ self.c)**3  / 1e+3

            self._masses = np.logspace(np.log10(m_min), np.log10(m_max), 5000)

            #precompute the halo mass function
            self._dndmh = dn_dm(self._z_array, self._masses, self.k, self.pk, omega_m = (0.3*(0.7**2)), h=0.7, sheth_a=self.sheth_a, sheth_q=self.sheth_q, sheth_p=self.sheth_p, window=self.window, c=self.c)

        super().__init__(parameters)


    @property
    def k(self):
        return self._k
    
    @property
    def pk(self):
        return self._pk



    def _loglkl(self, x, **kwargs):
     
        omega_b = x[:, self.index['Ombh2']]

        h       = np.full_like(omega_b, fill_value=0.7)
        omega_m = np.full_like(omega_b, fill_value=0.3 * (0.7**2))


        alpha_star = x[:, self.index['ALPHA_STAR']]
        t_star     = x[:, self.index['t_STAR']]
        f_star10   = 10**x[:, self.index['F_STAR10']]
        m_turn     = 10**x[:, self.index['M_TURN']]
     
        log_lkl = np.zeros(x.shape[0])

        # if we pass pk in argument then we cannot use 
        # the precoputed values for the halo mass function
        # (that is self._dndmh)
        precompute = self._precompute
        if kwargs.get('pk', None) is not None:
            precompute = False

        # allow k and pk to be passed as extra arguments of the function
        # in the case they depend on some parameters in theta
        k  = kwargs.get('k', self.k)
        pk = kwargs.get('pk', self.pk)

        if len(k.shape) == 1:  k  = np.tile(k, (omega_b.shape[0], 1))
        if len(pk.shape) == 1: pk = np.tile(pk, (omega_b.shape[0], 1))

        # loop on the datasets
        # we do not include Bouwens et al 2015 (10.1088/0004-637X/803/1/34)
        # stored at index 0, therefore we start the loop at 1
        for j in [1, 2, 3]:

            # loop on the redshift bins
            for iz, z, in enumerate(self.z_uv_exp[j]):

                hz = 100 * h_factor_no_rad(z, omega_b, omega_m - omega_b, h) * CONVERSIONS.km_to_mpc
                mh, mask_mh = m_halo(hz, self.m_uv_exp[j][iz], alpha_star, t_star, f_star10, omega_b, omega_m)

                try:

                    if precompute is False:
                        # predict the UV luminosity function on the range of magnitude m_uv at that redshift bin
                        # in the future, could add sheth_a, sheth_q, sheth_p and c as nuisance parameters
                        phi_uv_pred_z = phi_uv(z, hz, self.m_uv_exp[j][iz], k, pk, alpha_star, t_star, f_star10, m_turn, omega_b, omega_m, h, 
                                                    self.sheth_a, self.sheth_q, self.sheth_p, window = self.window, c = self.c, mh = mh, mask = mask_mh)
                        

                    else:

                        dndmh = interpolate.interp1d(self._masses, self._dndmh[0, self._z_index_from_table[j][iz], :])(mh)
                        phi_uv_pred_z = phi_uv(z, hz, self.m_uv_exp[j][iz], k, pk, alpha_star, t_star, f_star10, m_turn, omega_b, omega_m, h, 
                                                    self.sheth_a, self.sheth_q, self.sheth_p, window = self.window, c = self.c, mh = mh, 
                                                    mask = mask_mh, dndmh = dndmh)    
                    
                    phi_uv_pred_z = np.squeeze(phi_uv_pred_z, axis=1)

                    #if np.any(np.isnan(phi_uv_pred_z.flatten())):
                        #print(z, hz, self.m_uv_exp[j][iz], k, pk, alpha_star, t_star, f_star10, m_turn, omega_b, omega_m, h, mh, mask_mh)
                        #raise ValueError('Arghhh')
                    
                except ShortPowerSpectrumRange:
                    # kill the log likelihood in that case by setting it to -infinity
                    raise ValueError('Power spectrum not evaluated on a large enough range')               

                # get a sigma that is either the down or the up one depending 
                # if prediction is lower / higher than the observed value        
                mask         = (phi_uv_pred_z > self.phi_uv_exp[j][iz][None, :])
                sigma        = np.tile(self.sigma_phi_uv_down[j][iz], (x.shape[0], 1))
                sigma[mask]  = np.tile(self.sigma_phi_uv_up[j][iz],  (x.shape[0], 1))[mask]
        
                # update the log likelihood
                contrib = np.sum(np.log(np.sqrt(2.0/np.pi)/(self.sigma_phi_uv_up[j][iz] + self.sigma_phi_uv_down[j][iz])) - (phi_uv_pred_z - self.phi_uv_exp[j][iz])**2/(2*(sigma**2)), axis=-1)
                log_lkl = log_lkl + contrib

        return log_lkl
    


    def get_k_max(self, x) -> (None | float):
        
    
        omega_b = x[:, self.index['Ombh2']]

        h       = np.full_like(omega_b, fill_value=0.7)
        omega_m = np.full_like(omega_b, fill_value=0.3 * (0.7**2))

        alpha_star = x[:, self.index['ALPHA_STAR']]
        t_star     = x[:, self.index['t_STAR']]
        f_star10   = 10**x[:, self.index['F_STAR10']]

        min_mh = np.full(x.shape[0], fill_value=np.inf)

        for j in [1, 2, 3]:

            # loop on the redshift bins
            for iz, z, in enumerate(self.z_uv_exp[j]):

                hz = 100 * h_factor_no_rad(z, omega_b, omega_m - omega_b, h) * CONVERSIONS.km_to_mpc # approximation of the hubble factor
                mh, _ = m_halo(hz, self.m_uv_exp[j][iz], alpha_star, t_star, f_star10, omega_b, omega_m)

                mh = np.squeeze(mh, axis=1)   
                min_mh = np.minimum(min_mh, np.min(mh, axis=-1))
                    
        rhom0  = omega_m * CST_MSOL_MPC.rho_c_over_h2        
        k_max = 1.3 * self.c * (3*min_mh/(4*np.pi)/rhom0)**(-1/3)

        # one should (almost) never need self.kmax if large enough
        # set here as a security to do not make CLASS take to much
        # time and crash
        # print("Value of k_max: ", np.minimum(k_max / h, 10000))
        return np.minimum(k_max / h, 10000)





class OpticalDepthLikelihood(Likelihood):

    def __init__(self, 
                 parameters: list[str],  
                 *,
                 classifier: Classifier,
                 regressor:  Regressor,
                 median_tau:   float  = 0.0544,
                 sigma_tau:  float = 0.0073) -> None:
        
        self._classifier = classifier
        self._regressor  = regressor 
        self._median_tau   = median_tau
        self._sigma_tau  = sigma_tau   

        if not isinstance(self._sigma_tau, np.ndarray):
            self._sigma_tau = np.array([self._sigma_tau])

        super().__init__(parameters)

        # define an 'order list' to reorganise the parameters
        # in the order they are passed to the classifier and regressor
        ordered_params = regressor.metadata.parameters_name
        self._order    = [self.parameters.index(param) for param in ordered_params if param in self.parameters]

        
    @property
    def classifier(self):
        return self._classifier
    
    @property
    def regressor(self):
        return self._regressor
    
    @property
    def median_tau(self):
        return self._median_tau
    
    @property
    def sigma_tau(self):
        return self._sigma_tau


    def _loglkl(self, x, **kwargs):

        # get the number of parallel evaluations
        n = x.shape[0]

        xx = np.array([DEFAULT_VALUES[param] for param in self.regressor.metadata.parameters_name])
        xx = np.tile(xx, (n, 1))

        indices = np.array([list(self.regressor.metadata.parameters_name).index(param) for param in self.regressor.metadata.parameters_name if param in self.parameters])
        
        xx[:, indices] = x[:, self._order]

        # predict the ionization fraction from the NN
        xHII = predict_xHII_numpy(xx, self.classifier, self.regressor)

        # setting the result to -inf when the classifier returns it as a wrong value
        #loglkl = np.zeros(n)
        #loglkl[xHII[:, 0] == -1] = -np.inf
        # get the values in input (if given) or initialise to Planck 2018 results
        #tau     = self.mean_tau
        #var_tau = self.sigma_tau**2
        #loglkl = loglkl - 0.5 * ((tau- tau_pred)**2/var_tau + np.log( 2*np.pi * var_tau))

        # setting the result to -inf when the classifier returns it as a wrong value
        # initialise the result
        res = np.zeros(n)
        res[xHII[:, 0] == -1] = -np.inf

        # compute the optical depth to reionization
        tau_pred = predict_tau_from_xHII_numpy(xHII, xx, self.regressor.metadata)
        
        s_tau_m = np.full(n, fill_value=self.sigma_tau[0])
        s_tau_p = np.full(n, fill_value=self.sigma_tau[1])
        m_tau   = np.full(n, fill_value=self.median_tau) - s_tau_p * np.sqrt(2.0) * special.erfcinv(0.5*(1+s_tau_m/s_tau_p))

        # compute the truncated gaussian for the reionization data
        mask  = tau_pred > m_tau
        s_tau = s_tau_m
        s_tau[mask] = s_tau_p[mask]

        res = res + np.log(np.sqrt(2.0/np.pi)/(s_tau_m + s_tau_p)) - (tau_pred - m_tau)**2/(2*s_tau**2)

        return res
    
    


class ReionizationLikelihood(Likelihood):

    def __init__(self, 
                 parameters: list[str],
                 *,
                classifier: Classifier,
                regressor:  Regressor):
        
                
        self._classifier = classifier
        self._regressor  = regressor 

        # defining McGreer data
        # https://doi.org/10.1093/mnras/stu2449
        self.z_reio     = np.array([5.6, 5.9])
        self.x_reio     = np.array([0.96, 0.94])
        self.std_x_reio = np.array([0.05, 0.05])

        super().__init__(parameters)

        # define an 'order list' to reorganise the parameters
        # in the order they are passed to the classifier and regressor
        ordered_params = regressor.metadata.parameters_name
        self._order    = [self.parameters.index(param) for param in ordered_params if param in self.parameters]
        
       

    @property
    def classifier(self):
        return self._classifier
    
    @property
    def regressor(self):
        return self._regressor
    



    def _loglkl(self, x, **kwargs):

        # get the number of parallel evaluations
        n = x.shape[0]

        xx = np.array([DEFAULT_VALUES[param] for param in self.regressor.metadata.parameters_name])
        xx = np.tile(xx, (n, 1))

        indices = np.array([list(self.regressor.metadata.parameters_name).index(param) for param in self.regressor.metadata.parameters_name if param in self.parameters])
        
        xx[:, indices] = x[:, self._order]

        # predict the ionization fraction from the NN
        xHII_full = predict_xHII_numpy(xx, self.classifier, self.regressor)
        xHII = interpolate.interp1d(self.regressor.z, xHII_full)(self.z_reio)

        # setting the result to -inf when the classifier returns it as a wrong value
        # initialise the result
        res = np.zeros((n, *self.z_reio.shape))
        res[xHII[:, 0] == -1] = -np.inf
        
        std = np.tile(self.std_x_reio, (n, 1))
        x_reio = np.tile(self.x_reio, (n, 1))

        # compute the truncated gaussian for the reionization data
        norm_reio = -np.log(1.0 - x_reio + np.sqrt(np.pi/2.0)*std*special.erf(x_reio/(np.sqrt(2))/std))
        res = res + norm_reio

        mask = xHII < x_reio
        

        res[mask] = res[mask] - 0.5 * (xHII[mask] - x_reio[mask])**2/(std[mask]**2)

        return np.sum(res, axis=-1)
    

    

##############
# MCMC related functions

def log_prior(theta: np.ndarray, 
              theta_min: np.ndarray, 
              theta_max: np.ndarray, 
              **kwargs) -> np.ndarray:
    
    """
    Natural logarithm of the prior

    assume flat prior except for the parameters for which
    a covariance matrix and average value are given

    Parameters:
    -----------
    - theta: (n, d) ndarray
        parameters
        d is the dimension of the vector parameter 
        n is the number of vector parameter treated at once
    - theta_min: (d) ndarray
        minimum value of the parameters allowed
    - theta_max:
        maximum value of the parameters allowed

    kwargs:
    -------
    - mask: optional, (d) ndarray
        where the covariance matrix applies
        the mask should have p Trues and d-p False
        with p the dimension of the covariance matrix
        if cov and my given with dim d then mask still optional
    - mu: optional, (p) ndarray
        average value of the gaussian distribution
    - cov: optional, (p, p) ndarray
        covariance matrix
    """

    if len(theta.shape) == 1:
        theta = theta[None, :]

    res = np.zeros(theta.shape[0])

    # setting unvalid values to -infinity
    res[np.any(theta < theta_min, axis=-1)] = -np.inf
    res[np.any(theta > theta_max, axis=-1)] = -np.inf

    cov:  np.ndarray = kwargs.get('cov',  None)
    mu:   np.ndarray = kwargs.get('mu',   None)
    mask: np.ndarray = kwargs.get('mask', None)
    
    # add a gaussian distribution from a covariance matrix
    if (cov is not None) and (mu is not None):

        p = mu.shape[0]
        d = theta.shape[-1]

        # first makes some sanity checks
        assert cov.shape[0] == p, "Incompatible dimensions of the average vector and covariance matrix"
        if (mask is None) and (p == d):
            mask = np.full(d, fill_value=True, dtype=bool)

        inv_cov = np.linalg.inv(cov)
        det_cov = np.linalg.det(cov)

        # we perform the multiplicate of (p, p) with (n, p) on axes (1, 1) and then of (p, n) with (p, n) on axes (0, 0) and get back the diagonal value of the resulting (n, n)
        numerator   = - 0.5 * np.diagonal(np.tensordot((theta[:, mask] - mu).T,  np.tensordot(inv_cov, (theta[:, mask] - mu), axes=(1, 1)), axes=(0, 0)))
        denominator = + 0.5 * ( p * np.log(2*np.pi) + np.log(det_cov))
    
        res = res + numerator - denominator

    return res  


def log_likelihood(theta: np.ndarray, 
                   xi: np.ndarray,
                   likelihoods: list[Likelihood],
                   **kwargs) -> np.ndarray:
    """
    Compute the log Likelihood values.

    Parameters
    ----------
    theta : np.ndarray
        Varying parameters.
    xi : np.ndarray
        Extra fixed parameters.
    likelihoods : list[Likelihood]
        The likelihoods to evaluate for the fit.

    Returns
    -------
    np.ndarray
        Values of the log Likelihood for each chain.
    """
    
    res = np.zeros(theta.shape[0])

    need_pk: bool   = False
    uvlf_lkl: UVLFLikelihood | None = None

    if 'matter_power_spectrum_computer' in kwargs:
        for likelihood in likelihoods:
            if isinstance(likelihood, UVLFLikelihood):
                need_pk  = True
                uvlf_lkl = likelihood

    if need_pk is True:

        # first, get the k_max needed
        x: np.ndarray      = uvlf_lkl.get_x(theta, xi)
        k_max: float       = uvlf_lkl.get_k_max(x)

        # second, compute the matter power spectrum from a given function
        k_arr: np.ndarray  = np.logspace(np.log10(uvlf_lkl.k[0]), np.log10(k_max), 50000).T
        func: Callable     = kwargs.get('matter_power_spectrum_computer')
        x_dicts: dict       = uvlf_lkl.get_x_dicts(x)

        pk_arr = np.zeros((x.shape[0], k_arr.shape[1]))

        for i in range(x.shape[0]):
            pk_arr[i] = func(k_arr[i], x_dicts[i])

        kwargs['k']  = k_arr
        kwargs['pk'] = pk_arr
        
    # makes the sum of the log prior and log likelihood 
    for likelihood in likelihoods:
        res = res  + likelihood.loglkl(theta, xi, **kwargs)

    return res


def log_probability(theta:     np.ndarray, 
                    xi:        np.ndarray,
                    theta_min: np.ndarray, 
                    theta_max: np.ndarray,
                    likelihoods: list[Likelihood],
                    **kwargs) -> np.ndarray:
    
    if len(theta.shape) == 1:
        theta = theta[None, :]

    # compute the log prior
    res = log_prior(theta, theta_min, theta_max, **kwargs)
    
    # mask the infinities as we cannot compute the log_likelihood there
    mask = np.isfinite(res)
    res[~mask] = -np.inf

    res[mask] = log_likelihood(theta[mask, :], xi, likelihoods, **kwargs)

    return res 


def initialise_walkers(theta_min: np.ndarray, 
                       theta_max: np.ndarray, 
                       xi       : np.ndarray,
                       likelihoods: list[Likelihood],
                       n_walkers: int = 64, 
                       **kwargs):
    
    i = 0
    n_params = theta_min.shape[0]
    pos      = np.zeros((0, n_params))
    
    while pos.shape[0] < n_walkers and i < 1000 * n_walkers:

        # draw a value for the initial position
        prop = uniform_to_true(np.random.rand(1, n_params), theta_min, theta_max)

        try:
            
            log_lkl = log_likelihood(prop, xi, likelihoods, **kwargs)
        
            # check that the likelihood is finite at that position
            if np.all(np.isfinite(log_lkl)):

                # if finite add it to the list of positions
                pos = np.vstack((pos, prop))

        except:
            pass

        
        i = i+1

    if i >= 1000 * n_walkers:
        warnings.warn("The initialisation hit the safety limit of 1000 * n_walkers to initialise, pos may not be of size n_walkers.\n\
                      Consider reducing the parameter range to one where the likelihood is proportionnaly more defined")
        
    return pos






##################################################################################
# This file is part of NNERO.
#
# Copyright (c) 2024, Gaétan Facchinetti
#
# NNERO is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. NNERO is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with NNERO. 
# If not, see <https://www.gnu.org/licenses/>.
#
##################################################################################

import numpy as np
import torch
import torch.nn as nn

from .data    import TorchDataset, DataSet
from .network import NeuralNetwork

import os
import pkg_resources

from typing import Self

DATA_PATH = pkg_resources.resource_filename('nnero', 'nn_data/')

class Classifier(NeuralNetwork):
    """
    Daughter class of :py:class:`NeuralNetwork` specialised for classifier.
    
    Parameters
    ----------
    model: torch.nn.Module | None
        If not None, the model that will be used for classifier. 
        Otherwise, a new model is constructed from `n_input`, `n_hidden_features` 
        and `n_hidden_layers`. Default is None.
    n_input: int, optional
        Number of input on the neural network 
        (corresponds to the number of parameters).
        Default is 16.
    n_hidden_features: int, optional
        Number of hidden features per layer. Default is 32.
    n_hidden_layers: int, optional
        Number of layers. Default is 4.
    name: str | None
        Name of the neural network. If None, automatically set to DefaultClassifier.
        Default is None.
    dataset: Dataset | None
        Dataset on which the model will be trained. 
        If provided, gets `n_input` from the data and overrides the user input value.

    Attributes
    ----------
    - name : str
        the name of the model
    """

    def __init__(self, 
                *,
                n_input: int = 16, 
                n_hidden_features: int = 32, 
                n_hidden_layers: int = 4, 
                model: torch.nn.Module | None = None, 
                name: str | None = None,
                dataset: DataSet | None = None,
                dropouts: list[float] = []) -> None:

        # if no name, give a default
        if name is None:
            name = "DefaultClassifier"

        if dataset is not None:     
            n_input = len(dataset.metadata.parameters_name)


        # give a default empty array for the structure
        # stays None if a complex model is passed as input
        struct = np.empty(0)

        # initialise the default dropout layers
        if len(dropouts) > 0 and len(dropouts) != n_hidden_layers:
            raise ValueError("The number of dropouts values should be equal to the number of hidden layers")
    
        if len(dropouts) == 0:
            dropouts = [0.0] * n_hidden_layers
    
        # if no model defined in input give a model
        if model is None:
            
            # define a list of hidden layers
            hidden_layers = []
            for il in range(n_hidden_layers):
                
                hidden_layers.append(nn.Linear(n_hidden_features, n_hidden_features))
                hidden_layers.append(nn.ReLU())
                
                if dropouts[il] > 0:
                    hidden_layers.append(nn.Dropout(dropouts[il]))

            # create a sequential model
            model  = nn.Sequential(nn.Linear(n_input, n_hidden_features), *hidden_layers, nn.Linear(n_hidden_features, 1), nn.Sigmoid())
            
            # save the structure of this sequential model
            struct = np.array([n_input, n_hidden_features, n_hidden_layers, *dropouts])

        # call the (grand)parent init function
        super(Classifier, self).__init__(name)
        super(NeuralNetwork, self).__init__()

        # structure of the model
        self._struct = struct

        # define the model
        self._model = model

        # define the loss function (here binary cross-entropy)
        self._loss_fn = nn.BCELoss()

        # if the dataset is already given, set it as the dataset of the network
        if dataset is not None:
            self.set_check_metadata_and_partition(dataset)



    @classmethod
    def load(cls, path: str | None = None) -> Self:
        """
        Loads a classifier.

        Parameters
        ----------
        path: str | None
            Path to the saved files containing the classifier data.
            If None automatically fetch the DefaultClassifier.

        Returns
        -------
        Classifier
        """

        if path is None: 
            path = os.path.join(DATA_PATH, "DefaultClassifier")

        name = path.split('/')[-1]
        
        if os.path.isfile(path  + '_struct.npy'):

            with open(path  + '_struct.npy', 'rb') as file:
                struct  = np.load(file)

                good_structure: bool = False

                if len(struct) == 3 + int(struct[2]):
                    
                    dropouts = []
                    for j in range(3, len(struct)):
                        dropouts.append(struct[j])

                    good_structure = True

                # for compatibility with older versions
                if len(struct) == 3:

                    dropouts = [0.0] * int(struct[2])
                    good_structure = True

                if good_structure: 
                    classifier = Classifier(n_input=int(struct[0]), n_hidden_features=int(struct[1]), n_hidden_layers=int(struct[2]), dropouts=dropouts)
                    classifier.load_weights_and_extras(path)
                    classifier.eval()

                    print('Model ' + str(name) + ' sucessfully loaded')

                    return classifier
        
        # if the struct read is not of the right size
        # check for a pickled save of the full class
        # (although this is not recommended)
        if os.path.isfile(path  + '.pth') :
            classifier = torch.load(path + ".pth", weights_only=False)
            classifier.eval()

            print('Model ' + str(name) + ' sucessfully loaded from a .pth archive')

            return classifier
        
        raise ValueError("Could not find a fully saved classifier model at: " + path)


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward evaluation of the model.

        Parameters
        ----------
        x: torch.Tensor
            Input features.
        """
        return torch.flatten(self._model(x))
    
    @property
    def loss_fn(self):
        return self._loss_fn
    
    def test(self, dataset: DataSet | None = None, x_test:np.ndarray | None = None, y_test: np.ndarray | None = None) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Test the efficiency of the classifier.

        Parameters
        ----------
        dataset: DataSet | None
            DataSet containing the training partition and the test partition.
        x_test: np.ndarray
        y_test: np.ndarray

        Returns
        -------
        tuple(np.ndarray, np.ndarray, np.ndarray)
            y_pred, y_test, and array of true if rightly classifier, false otherwise 

        Raises
        ------
        ValueError
            Either the dataset or both x_test and y_test must be provided.
        """

        if dataset is not None:
            self.set_check_metadata_and_partition(dataset, check_only = True)
            x_test = torch.tensor(dataset.x_array[dataset.partition.total_test],      dtype=torch.float32)
            y_test = torch.tensor(dataset.y_classifier[dataset.partition.total_test], dtype=torch.float32)
        elif (x_test is not None) and (y_test is not None):
            x_test = torch.tensor(x_test, dtype=torch.float32)
            y_test = torch.tensor(y_test, dtype=torch.float32)
        else:
            raise ValueError("Either the dataset or both x_test and y_test must be provided.")
        
        self.eval()
        
        with torch.no_grad():
            y_pred  = self.forward(x_test)
            print(f"The accuracy is {100*(y_pred.round() == y_test).float().mean():.4f}%")
            return y_pred.numpy(), y_test.numpy(), (y_pred.round() == y_test).numpy()
        

    def validate(self, dataset: DataSet) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Validate the efficiency of the classifier.

        Parameters
        ----------
        dataset: DataSet
            DataSet containing the training partition and the test partition.

        Returns
        -------
        tuple(np.ndarray, np.ndarray, np.ndarray)
            y_pred, y_test, and array of true if rightly classifier, false otherwise 
        """

        self.set_check_metadata_and_partition(dataset, check_only = True)
        x_valid = torch.tensor(dataset.x_array[dataset.partition.total_valid],      dtype=torch.float32)
        y_valid = torch.tensor(dataset.y_classifier[dataset.partition.total_valid], dtype=torch.float32)
        
        self.eval()
        
        with torch.no_grad():
            y_pred  = self.forward(x_valid)
            print(f"The accuracy is {100*(y_pred.round() == y_valid).float().mean():.4f}%")
            return y_pred.numpy(), y_valid.numpy(), (y_pred.round() == y_valid).numpy()


    
def train_classifier(model: Classifier, 
                     dataset: DataSet, 
                     optimizer:torch.optim.Optimizer, 
                     *, 
                     epochs: int = 50, 
                     learning_rate: float = 1e-3, 
                     verbose: bool = True, 
                     batch_size: int = 64, 
                     x_train: np.ndarray | None = None,
                     y_train: np.ndarray | None = None,
                     x_valid: np.ndarray | None = None,
                     y_valid: np.ndarray | None = None,
                     **kwargs)-> None:
    """
    Trains a given classifier.

    Parameters
    ----------
    model : Classifier
        Classifier model to train.
    dataset : DataSet
        Dataset on which to train the classifier.
    optimizer : torch.optim.Optimizer
        Optimizer used for training.
    epochs : int, optional
        Number of epochs, by default 50.
    learning_rate : float, optional
        Learning rate for training, by default 1e-3.
    verbose : bool, optional
        If true, outputs a summary of the losses at each epoch, by default True.
    batch_size : int, optional
        Size of the training batches, by default 64.
    """
    # set the metadata and parition object of the model
    model.set_check_metadata_and_partition(dataset)

    if x_train is None:
        x_train = dataset.x_array[dataset.partition.total_train]
    
    if y_train is None:
        y_train = dataset.y_classifier[dataset.partition.total_train]

    if x_valid is None:
        x_valid = dataset.x_array[dataset.partition.total_valid]

    if y_valid is None:
        y_valid = dataset.y_classifier[dataset.partition.total_valid]
        
    # format the data for the classifier
    train_dataset = TorchDataset(x_train, y_train)
    valid_dataset = TorchDataset(x_valid, y_valid)
    train_loader  = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size, shuffle=True, **kwargs)
    valid_loader  = torch.utils.data.DataLoader(valid_dataset, batch_size=batch_size, shuffle=False, **kwargs)
    
    # we have only one param_group here
    # we modify the learning rate of that group
    optimizer.param_groups[0]['lr'] = learning_rate

    # start loop on the epochs
    for epoch in range(epochs):
        
        train_loss     = np.array([])
        valid_loss     = np.array([])
        train_accuracy = np.array([])
        valid_accuracy = np.array([])

        # training mode
        model.train()
        
        for batch in train_loader:
            x_batch, y_batch = batch

            optimizer.zero_grad()
            y_pred = model.forward(x_batch)
            loss   = model.loss_fn(y_pred, y_batch)
            loss.backward()
            optimizer.step()
            
            train_loss     = np.append(train_loss, loss.item())
            train_accuracy = np.append(train_accuracy, (y_pred.round() == y_batch).float().mean())


        # evaluation mode
        model.eval()
    
        with torch.no_grad():
            
            for batch in valid_loader:
                x_batch, y_batch = batch
            
                y_pred = model(x_batch)             # forward pass
                loss = model.loss_fn(y_pred, y_batch) # loss function averaged over the batch size
                
                valid_loss     = np.append(valid_loss, loss.item())
                valid_accuracy = np.append(valid_accuracy, (y_pred.round() == y_batch).float().mean())
        
        # get the mean of all batches
        model._train_loss     = np.append(model._train_loss, np.mean(train_loss))
        model._valid_loss     = np.append(model._valid_loss, np.mean(valid_loss))
        model._train_accuracy = np.append(model._train_accuracy, np.mean(train_accuracy))
        model._valid_accuracy = np.append(model._valid_accuracy, np.mean(valid_accuracy))

        if verbose:
            print(f'Epoch [{epoch+1}/{epochs}], loss: ({model.train_loss[-1]:.4f}, {model.valid_loss[-1]:.4f}), accuracy = ({model.train_accuracy[-1]:.4f}, {model.valid_accuracy[-1]:.4f})')   




##################################################################################
# This file is part of NNERO.
#
# Copyright (c) 2024, Gaétan Facchinetti
#
# NNERO is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. NNERO is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with NNERO. 
# If not, see <https://www.gnu.org/licenses/>.
#
##################################################################################


import numpy as np
import torch
import torch.nn as nn

from .data      import TorchDataset, DataSet, uniform_to_true
from .network   import NeuralNetwork
from .cosmology import optical_depth_no_rad

import os
import pkg_resources

DATA_PATH = pkg_resources.resource_filename('nnero', 'nn_data/')


class Regressor(NeuralNetwork):
    """
    Daughter class of :py:class:`NeuralNetwork` specialised for regressors.
    
    Parameters
    ----------
    model: torch.nn.Module | None
        If not None, the model that will be used for classifier. 
        Otherwise, a new model is constructed from `n_input`, `n_hidden_features` 
        and `n_hidden_layers`. Default is None.
    n_input: int, optional
        Number of input on the neural network 
        (corresponds to the number of parameters).
        Default is 16.
    n_input: int, optional
        Number of output on the neural network 
        (corresponds to the number of redshift bins).
        Default is 50. Overriden if `dataset` is specified.
    n_hidden_features: int, optional
        Number of hidden features per layer. Default is 80.
    n_hidden_layers: int, optional
        Number of layers. Default is 5.
    name: str | None
        Name of the neural network. If None, automatically set to DefaultClassifier.
        Default is None.
    dataset: Dataset | None
        Dataset on which the model will be trained. 
        If provided, gets `n_input` and `n_output` from the data and overrides the user input value.
    use_pca: bool, optional
        If `True`, decompose the interpolated function on the principal component eigenbasis.
        Default is True.
    pca_precision: float, optional
        If `use_pca` is `True` sets how many eigenvalues needs to be considered.
        Only consider eigenvectors with eigenvalues > precision^2 * max(eigenvalues).
        Default is 1e-3.
    alpha_tau: float, optioal
        Weighting of the relative error on X_e and optical depth in the cost function.
        Default is 0.5
        
    Attributes
    ----------
    - name : str
        the name of the model
    """

    def __init__(self, 
                 *, 
                 n_input: int = 16, 
                 n_output: int = 50, 
                 n_hidden_features: int = 80, 
                 n_hidden_layers: int = 5, 
                 alpha_tau : float = 0.5,
                 model = None, 
                 name: str | None = None,
                 use_pca: bool = True,
                 pca_precision: float = 1e-3,
                 dataset: DataSet | None = None):

        if name is None:
            name = "DefaultRegressor"

        # save as attribute wether or not we want to work
        # in the principal component eigenbasis
        self._use_pca = use_pca
        self._pca_precision = pca_precision
        
        # if we provide a dataset with pca_precision we override
        # the number of output to correspond to the number of
        # usefull eigenvectors otherwise it is set to the number 
        # of redshift bins used 
        # if no dataset is given, need to know in advance how
        # many usefull eigenvectors there are and pass is as n_output
        # ----
        # this also initialises the pca vectors in the metadata 
        # attribute of the dataset object
        # ----
        # when loading from a file we do not provide a dataset
        # and therefore do not redo the initialisation of the pca,
        # pca eigenvectors and all are read from the saved metadata
        if dataset is not None:
            
            n_input = len(dataset.metadata.parameters_name)

            if use_pca is True:
                n_output = dataset.init_principal_components(self.pca_precision)
            else:
                n_output = len(dataset.metadata.z)
            
            
        # give a default empty array for the structure
        # stays None if a complex model is passed as input
        struct = np.empty(0)

        # definition of the model if none is given as input
        if model is None:
            
            hidden_layers = []
            for _ in range(n_hidden_layers):
                hidden_layers.append(nn.Linear(n_hidden_features, n_hidden_features))
                hidden_layers.append(nn.ReLU())

            # create a sequential model
            model = nn.Sequential(nn.Linear(n_input, n_hidden_features), *hidden_layers, nn.Linear(n_hidden_features, n_output))
        
            # save the structure of this sequential model and more
            struct = np.array([n_input, n_output, n_hidden_features, n_hidden_layers, alpha_tau, use_pca, pca_precision])
        
        # call the (grand)parent constructors
        super(Regressor, self).__init__(name)
        super(NeuralNetwork, self).__init__()

        # structure of the model
        self._struct = struct
        
        # define the model
        self._model = model

        # define parameters in the loss
        self._alpha_tau = alpha_tau

        # if the dataset is already given, set it as the dataset of the network
        if dataset is not None:
            self.set_check_metadata_and_partition(dataset)

    


    @classmethod
    def load(cls, path: str | None = None):
        """
        Loads a regressor.

        Parameters
        ----------
        path: str | None
            Path to the saved files containing the regressor data.
            If None automatically fetch the DefaultRegressor.

        Returns
        -------
        Regressor
        """

        if path is None:
            path = os.path.join(DATA_PATH, "DefaultRegressor")

        name = path.split('/')[-1]
        
        if os.path.isfile(path  + '_struct.npy'):

            with open(path  + '_struct.npy', 'rb') as file:
                struct  = np.load(file)

                if len(struct) == 7:

                    regressor = Regressor(n_input=int(struct[0]), 
                                          n_output=int(struct[1]), 
                                          n_hidden_features=int(struct[2]), 
                                          n_hidden_layers=int(struct[3]),
                                          alpha_tau=struct[4],
                                          use_pca=bool(struct[5]),
                                          pca_precision=struct[6],
                                          name=name)
                    regressor.load_weights_and_extras(path)
                    regressor.eval()

                    print('Model ' + str(name) + ' sucessfully loaded')

                    return regressor
        
        # if the struct read is not of the right size
        # check for a pickled save of the full class
        # (although this is not recommended)
        if os.path.isfile(path  + '.pth') :
            regressor = torch.load(path + ".pth")
            regressor.eval()

            print('Model ' + str(name) + ' sucessfully loaded from a .pth archive')
            
            return regressor
        
        raise ValueError("Could not find a fully saved regressor model at: " + path)

        
    def forward(self, x):
        """
        Forward evaluation of the model.

        Parameters
        ----------
        x: torch.Tensor
            input features
        """

        if self.use_pca is True:
            # reconstruct the value of the function in the principal component eigenbasis
            y = torch.matmul(self._model(x), self.metadata.torch_pca_eigenvectors[0:self.metadata.pca_n_eigenvectors, :])
            y = y + self.metadata.torch_pca_mean_values
            y = 10**y
        else:
            y = self._model(x)
        return torch.clamp(y, max=1.0)
    
    def tau_ion(self, x, y):
        """
        Optical depth to reionization.

        Parameters
        ----------
        x: torch.Tensor
            Input features.
        y: torch.Tensor
            Output of the Regressor. Corresponds to X_e(z).
        """

        z_tensor = torch.tensor(self.metadata.z, dtype=torch.float32)
        omega_b  = uniform_to_true(x[:, self.metadata.pos_omega_b], self.metadata.min_omega_b, self.metadata.max_omega_b)
        omega_c  = uniform_to_true(x[:, self.metadata.pos_omega_dm], self.metadata.min_omega_c, self.metadata.max_omega_c)
        hlittle  = uniform_to_true(x[:, self.metadata.pos_hlittle], self.metadata.min_hlittle, self.metadata.max_hlittle)
        return optical_depth_no_rad(z_tensor, y, omega_b, omega_c, hlittle)
    
    def loss_xHII(self, output, target):
        return torch.mean(torch.abs(1.0-torch.div(output, target[:, :-1])))

    def loss_tau(self, tau_pred, target):
        return torch.mean(torch.abs(1.0-torch.div(tau_pred, target[:, -1])))
    
    def test_tau(self, dataset:DataSet) -> np.ndarray:
        """
        Test the efficiency of the regressor to reconstruct the optical depth to reionization.

        Parameters
        ----------
        dataset : DataSet
            DataSet containing the training partition and the test partition.

        Returns
        -------
        np.ndarray
            Distribution of relative error between the predicted and true optical depth. 
            Array with the size of the test dataset.
        """

        self.set_check_metadata_and_partition(dataset, check_only = True)
        x_test   = torch.tensor(dataset.x_array[dataset.partition.selection_test],     dtype=torch.float32)
        tau_test = torch.tensor(dataset.y_regressor[dataset.partition.selection_test, -1], dtype=torch.float32)
        
        self.eval()
        
        with torch.no_grad():
            y_pred   = self.forward(x_test)
            tau_pred = self.tau_ion(x_test, y_pred)

        return (1.0-tau_pred/tau_test).numpy()

    def test_xHII(self, dataset:DataSet) -> tuple[np.ndarray, np.ndarray]:
        """
        Test the efficiency of the regressor to reconstruct the free electron fraction X_e.

        Parameters
        ----------
        dataset : DataSet
            DataSet containing the training partition and the test partition.

        Returns
        -------
        tuple(np.ndarray, np.ndarray)
            Prediction for X_e and true values.
        """

        self.set_check_metadata_and_partition(dataset, check_only = True)
        x_test = torch.tensor(dataset.x_array[dataset.partition.selection_test],          dtype=torch.float32)
        y_test = torch.tensor(dataset.y_regressor[dataset.partition.selection_test, :-1], dtype=torch.float32)
        
        self.eval()
        
        with torch.no_grad():
            y_pred = self.forward(x_test)

        return y_pred.numpy(), y_test.numpy()
    
    @property
    def alpha_tau(self):
        return self._alpha_tau

    @property
    def use_pca(self):
        return self._use_pca
    
    @property
    def pca_precision(self):
        return self._pca_precision
    
    @property
    def z(self):
        return self._metadata.z
    

    
def train_regressor(model: Regressor, 
                    dataset: DataSet, 
                    optimizer: torch.optim.Optimizer, 
                    *, 
                    epochs = 50, 
                    learning_rate = 1e-3, 
                    verbose = True, 
                    batch_size = 64, 
                    **kwargs):
    """
    Trains a given regressor.

    Parameters
    ----------
    model : Regressor
        Regressor model to train.
    dataset : DataSet
        Dataset on which to train the regressor.
    optimizer : torch.optim.Optimizer
        Optimizer used for training.
    epochs : int, optional
        Number of epochs, by default 50.
    learning_rate : float, optional
        Learning rate for training, by default 1e-3.
    verbose : bool, optional
        If true, outputs a summary of the losses at each epoch, by default True.
    batch_size : int, optional
        Size of the training batches, by default 64.
    """
    
    # if we use pca but have not yet initialised it in metadata do it at first training step
    # this can happen if no dataset is given to initialise the model while havind use_pca = True
    if (len(model.train_loss) == 0) and (model.use_pca is True) and (len(dataset.metadata.pca_eigenvalues) == 0):
        dataset.init_principal_components(model.pca_precision)

    # set or check the metadata and parition attributes of the model
    model.set_check_metadata_and_partition(dataset)

    # format the data for the regressor
    train_dataset = TorchDataset(dataset.x_array[dataset.partition.selection_train], dataset.y_regressor[dataset.partition.selection_train])
    valid_dataset = TorchDataset(dataset.x_array[dataset.partition.selection_valid], dataset.y_regressor[dataset.partition.selection_valid])
    train_loader  = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size, shuffle=True, **kwargs)
    valid_loader  = torch.utils.data.DataLoader(valid_dataset, batch_size=batch_size, shuffle=False, **kwargs)

    # we have only one param_group here
    # we modify the learning rate of that group
    optimizer.param_groups[0]['lr'] = learning_rate

    # start loop on the epochs
    for epoch in range(epochs):
        
        train_loss     = np.array([])
        valid_loss     = np.array([])
        train_accuracy = np.array([])
        valid_accuracy = np.array([])

        # training mode
        model.train()
        
        for batch in train_loader:
            x_batch, y_batch = batch

            optimizer.zero_grad()
            y_pred   = model.forward(x_batch)
            tau_pred = model.tau_ion(x_batch, y_pred)
            
            loss_xHII = model.loss_xHII(y_pred, y_batch)
            loss_tau  = model.loss_tau(tau_pred, y_batch)
            loss     = (1.0-model.alpha_tau) * loss_xHII + model.alpha_tau * loss_tau

            loss.backward()
            optimizer.step()
            
            train_loss     = np.append(train_loss, loss.item())
            train_accuracy = np.append(train_accuracy, 1-loss_tau.item())


        # evaluation mode
        model.eval()
    
        with torch.no_grad():
            
            for batch in valid_loader:
                x_batch, y_batch = batch
            
                y_pred = model.forward(x_batch)              # forward pass
                tau_pred = model.tau_ion(x_batch, y_pred)
                
                loss_xHII = model.loss_xHII(y_pred, y_batch)
                loss_tau  = model.loss_tau(tau_pred, y_batch)
                loss      = (1.0-model.alpha_tau) * loss_xHII + model.alpha_tau * loss_tau
                
                valid_loss     = np.append(valid_loss, loss.item())
                valid_accuracy = np.append(valid_accuracy, 1-loss_tau.item())
        
        # get the mean of all batches
        model._train_loss     = np.append(model._train_loss, np.mean(train_loss))
        model._valid_loss     = np.append(model._valid_loss, np.mean(valid_loss))
        model._train_accuracy = np.append(model._train_accuracy, np.mean(train_accuracy))
        model._valid_accuracy = np.append(model._valid_accuracy, np.mean(valid_accuracy))

        if verbose:
            print(f'Epoch [{epoch+1}/{epochs}], loss: ({model.train_loss[-1]:.4f}, {model.valid_loss[-1]:.4f}), accuracy = ({model.train_accuracy[-1]:.4f}, {model.valid_accuracy[-1]:.4f})')   





##################################################################################
# This file is part of NNERO.
#
# Copyright (c) 2024, Gaétan Facchinetti
#
# NNERO is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License, or any 
# later version. NNERO is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU 
# General Public License along with NNERO. 
# If not, see <https://www.gnu.org/licenses/>.
#
##################################################################################

##################
#
# Definition of cosmological quantities
# Fast computation of the optical depth to reionization
# (heavily use of numpy or torch vectorisation)
# The cosmology is assumed to be flat (omega_k = 0)
#
##################

import numpy as np
import torch

from .constants import CST_EV_M_S_K, CST_NO_DIM, CST_MSOL_MPC, CONVERSIONS


def convert_array(arr: float, to_torch: bool = False) -> (np.ndarray | torch.Tensor):
    if isinstance(arr, np.ndarray | torch.Tensor):
        if (isinstance(arr, np.ndarray) and to_torch is False) or (isinstance(arr, torch.Tensor) and to_torch is True):
            return arr
        if isinstance(arr, np.ndarray) and to_torch is True:
            return torch.tensor(arr)
        if isinstance(arr, torch.Tensor) and to_torch is False:
            return arr.numpy()
    return np.array([arr]) if not to_torch else torch.tensor([arr])

 

#################################
## Simple functions that can be evaluated fast with numpy arrays

# ----------------------------------------------------------
# Densities and abundances

def rho_baryons(omega_b : float | np.ndarray | torch.Tensor) -> (float | np.ndarray | torch.Tensor):
    """
    Baryon energy density (in eV / m^3)

    Parameters
    ----------
    omega_b: float | np.ndarray | torch.Tensor
        reduced abundance of baryons (i.e. times h^2)

    Returns
    -------
    float or np.ndarray or torch.tensor
    """

    return  omega_b * CST_EV_M_S_K.rho_c_over_h2 # in eV / m^3


def n_baryons(omega_b : float | np.ndarray | torch.Tensor) -> (float | np.ndarray | torch.Tensor):
    """
    Baryon number density (in 1 / m^3)

    Parameters
    ----------
    omega_b: float | np.ndarray | torch.Tensor
        reduced abundance of baryons (i.e. times h^2)

    Returns
    -------
        float or np.ndarray or torch.tensor
    """

    return rho_baryons(omega_b) * ( (1.0- CST_NO_DIM.YHe) / CST_EV_M_S_K.mass_hydrogen +  CST_NO_DIM.YHe / CST_EV_M_S_K.mass_helium  ) # in 1/m^3


def n_hydrogen(omega_b : float | np.ndarray | torch.Tensor) -> (float | np.ndarray | torch.Tensor):
    """
    Hydrogen number density (in 1 / m^3)

    Parameters
    ----------
    omega_b: float | np.ndarray | torch.Tensor
        reduced abundance of baryons (i.e. times h^2)

    Returns
    -------
        float or np.ndarray or torch.tensor
    """

    return rho_baryons(omega_b) * (1.0 - CST_NO_DIM.YHe) / CST_EV_M_S_K.mass_hydrogen


def n_ur(m_nus: np.ndarray | torch.Tensor) -> (np.ndarray | torch.Tensor):
    """
    Number of ultra-relativistic degrees of freedom
    
    Parameters
    ----------
    m_nus: np.ndarray | torch.Tensor
        shape (q1, q2, ..., qn, 3), mass of the three neutrinos
        in a given model
    
    Returns
    -------
    np.ndarray or torch.Tensor with shape (q1, q2, ..., qn)
    """

    return CST_NO_DIM.Neff - np.count_nonzero(m_nus, axis=-1)


def omega_r(m_nus: np.ndarray | torch.Tensor) -> (np.ndarray | torch.Tensor):
    """
    Reduced abundance of radiation today
    
    Parameters
    -----------
    m_nus: np.ndarray | torch.Tensor
        shape (q1, q2, ..., qn, 3), mass of the three neutrinos
        in a given model

    Returns
    -------
    np.ndarray or torch.Tensor with shape (q1, q2, ..., qn)
    """
    return  4.48162687719e-7 * CST_EV_M_S_K.T0**4 * (1.0 + 0.227107317660239 * n_ur(m_nus))


def _interp_omega_nu(y):
    """ interpolation function """
    # This interpolation formula is taken from the HYREC-2 code: https://github.com/nanoomlee/HYREC-2
    return (1.+0.317322*0.0457584*y**(3.47446+1.) + 2.05298*0.0457584*y**(3.47446-1.))/(1.+0.0457584*y**(3.47446))*(3.45e-8*(CST_EV_M_S_K.Tnu0**4))*5.6822*2


def omega_nu(z:     float | np.ndarray | torch.Tensor, 
             m_nus: np.ndarray | torch.Tensor
             )-> np.ndarray | torch.Tensor :
    """
    Efficient implementation of reduced neutrino abundance for numpy arrays. 
    Parameters must be floats, arrays or tensor with compatible dimensions. 
    
    Parameters
    ----------
    m_nus: np.ndarray | torch.Tensor
        shape (q1, q2, ..., qn, 3), mass of the three neutrinos
        in a given model
    z: float | numpy.ndarray | torch.Tensor
        shape (p, ) if array, redshift range

    Returns
    -------
    numpy.ndarray or torch.Tensor with shape (q1, ..., qn, p)
    """

    # convert z to an array if it is a float
    z = convert_array(z)

    # a is of shape (1, ..., p, 1)
    p = len(z.flatten())
    a = 1.0/(1.0+z.reshape((*tuple(1 for _ in m_nus.shape[:-1]), p, 1)))

    m_nus = m_nus.reshape((*m_nus.shape[:-1], 1, 3))
    
    # y is of shape (q1, q2, ..., qn, p, 3) 
    y = m_nus/CST_EV_M_S_K.k_Boltz/(CST_EV_M_S_K.Tnu0/a)
    
    # res is of shape (q1, q2, ..., qn, p, 3)
    res = _interp_omega_nu(y)

    # y[..., i] is of shape (q1, ..., qn, p)
    # ensure that we do not sum over 0 entries
    for i in range(3):
        res[y[..., i] == 0, i] = 0

    return np.sum(res, axis=-1)




# ----------------------------------------------------------
# Hubble factor and optical depth


def h_factor_numpy(z:       float | np.ndarray | torch.Tensor,
                   omega_b: float | np.ndarray | torch.Tensor,
                   omega_c: float | np.ndarray | torch.Tensor, 
                   h:       float | np.ndarray | torch.Tensor,
                   m_nus:   np.ndarray | torch.Tensor
                   ) -> (np.ndarray | torch.Tensor):
    """
    E(z) = H(z) / H0 with radiation

    Efficient evalutation of hubble rate parameters for numpy arrays 
    or torch tensors (also works with simple float ). In the first
    case the shape of these numpy arrays must be compatible (see below).
    
    Parameters
    ----------
    z: float | numpy.ndarray | torch.Tensor
        shape (p, ) if array, redshift range
    omega_b: float | numpy.ndarray | torch.Tensor
        shape (q1,..., qn) if array, reduced abundance of baryons (i.e. times h^2)
    omega_c: float | numpy.ndarray | torch.Tensor
        shape (q1,..., qn) if array, reduced abundance of dark matter (i.e. times h^2)
    h: float | numpy.ndarray | torch.Tensor
        shape (q1,..., qn) if array,, Hubble factor
    mnus: numpy.ndarray | torch.Tensor
        shape (q1,..., qn, 3) if array or at least shape (, 3), mass of the neutrinos
       
    Returns
    -------
    numpy.ndarray or torch.Tensor with shape (q1,..., qn, p)
    """

    # convert the input if they are just floats
    omega_b = convert_array(omega_b)
    omega_c = convert_array(omega_c)
    h       = convert_array(h)
    z       = convert_array(z)

    if len(m_nus.shape) == 1:
        m_nus = m_nus[None, :]
                   
    # a is of shape (1, ... p)
    p = len(z.flatten())
    a = 1.0/(1.0+z.reshape((*tuple(1 for _ in h.shape), p)))

    # h is of shape (q1, ..., qn, 1)
    _h = h[..., None]

    # omega values is of shape (q1, ..., qn, 1)
    m_omega_r  = omega_r(m_nus)[..., None]
    m_omega_nu = omega_nu(0, m_nus)
    m_omega_m  = (omega_b + omega_c)[..., None]

    m_omega_l  = (_h**2) - m_omega_m - m_omega_r - m_omega_nu

    # result is of shape (q1, ..., qn, p)
    return np.sqrt(m_omega_m / (a**3) + (m_omega_r + omega_nu(z, m_nus)) / (a**4) + m_omega_l) / _h



def h_factor_no_rad(z:       float | np.ndarray | torch.Tensor, 
                    omega_b: float | np.ndarray | torch.Tensor, 
                    omega_c: float | np.ndarray | torch.Tensor, 
                    h:       float | np.ndarray | torch.Tensor
                    ) -> (np.ndarray | torch.Tensor):
    """
    E(z) = H(z) / H0 without radiation

    Efficient evalutation of hubble rate parameters for numpy arrays 
    or torch tensors (also works with simple float ). In the first
    case the shape of these numpy arrays must be compatible (see below).
    
    Parameters
    ----------
    z: float | numpy.ndarray | torch.Tensor
        shape (p, ) if array, redshift range
    omega_b: float | numpy.ndarray | torch.Tensor
        shape (q1,..., qn) if array, reduced abundance of baryons (i.e. times h^2)
    omega_c: float | numpy.ndarray | torch.Tensor
        shape (q1,..., qn) if array, reduced abundance of dark matter (i.e. times h^2)
    h: float | numpy.ndarray | torch.Tensor
        shape (q1,..., qn) if array, Hubble factor

    Returns
    -------
    numpy.ndarray or torch.Tensor with shape (q1,..., qn, p)
    """

    xp = np
    to_torch = False
    if isinstance(z, torch.Tensor):
        xp = torch
        to_torch = True

    # convert the input if they are just floats
    omega_b = convert_array(omega_b, to_torch)
    omega_c = convert_array(omega_c, to_torch)
    h       = convert_array(h, to_torch)
    z       = convert_array(z, to_torch)
                   
    # a is of shape (1, ..., p)
    p = len(z.flatten())
    a = 1.0/(1.0+z.reshape((*tuple(1 for _ in h.shape), p)))

    # h is of shape (q1, ..., qn, 1)
    _h = h[..., None]

    # omega values is of shape (q1, ..., qn, 1)
    m_omega_m  = (omega_b + omega_c)[..., None]
    m_omega_l  = (_h**2) - m_omega_m 

    # result is of shape (q1, ..., qn, p)
    return xp.sqrt(m_omega_m / (a**3)  + m_omega_l) / _h




def optical_depth_no_rad(z:       float | np.ndarray | torch.Tensor, 
                         xHII:    np.ndarray | torch.Tensor,
                         omega_b: float | np.ndarray | torch.Tensor, 
                         omega_c: float | np.ndarray | torch.Tensor, 
                         h:       float | np.ndarray | torch.Tensor,
                         *, 
                         low_value: float = 1.0,
                         with_helium: bool = True,
                         cut_integral_min: bool = True):
    """
    Optical depth to reionization without radiation.

    Efficient evaluation of the opetical depth to reionization (dimensionless)
    uses fast numpy / torch operations with trapezoid rule
    (assume that radiation is neglibible on the range of z). Also neglegts the
    influence of double reionization of helium at small redshifts


    Parameters
    ----------
    z: float | numpy.ndarray | torch.Tensor
        shape (p,) if array, redshift range
    xHII: numpy.ndarray | torch.Tensor
        shape (n, p), ionization fraction vs the redshift for the n models
    omega_b: float | numpy.ndarray | torch.Tensor
        shape (n,) if array, reduced abundance of baryons (i.e. times h^2)
    omega_c: float | numpy.ndarray | torch.Tensor
        shape (n,) if array, reduced abundance of dark matter (i.e. times h^2)
    h: float | numpy.ndarray | torch.Tensor
        shape (n,) if array, Hubble factor
    low_value: float
        value of xHII at redshift smaller than min(z)


    Returns
    -------
    numpy.ndarray or torch.Tensor with shape(n, p)
    """
    
    # define if we work with torch or numpy
    xp = np
    to_torch = False
    if isinstance(xHII, torch.Tensor):
        xp = torch
        to_torch = True

    # convert the floats to arrays or tensors
    z       = convert_array(z, to_torch)
    h       = convert_array(h, to_torch)
    omega_b = convert_array(omega_b, to_torch)
    omega_c = convert_array(omega_c, to_torch)

    if len(z.shape) == 1:
        if isinstance(z, np.ndarray):
            z = z[None, :]
        if isinstance(z, torch.Tensor):
            z = z.unsqueeze(0)

    if len(xHII.shape) == 1:
        if isinstance(xHII, np.ndarray):
            xHII = xHII[None, :]
        if isinstance(xHII, torch.Tensor):
            xHII = xHII.unsqueeze(0)

    # prepare data for redshifts < min(z)
    # we assume that at small z value xHII = 1
    z_small    = xp.linspace(0,  z.min(), 20)[None, :]    
    xHII_small = xp.full((xHII.shape[0], z_small.shape[-1]), fill_value=low_value)

    # concatenate the two parts (small and large z)
    rs = xp.concatenate((z_small, z), axis=-1)
    Xe = xp.concatenate((xHII_small, xHII), axis=-1)

    # add the Helium contribution
    if with_helium is True:
        fHe = CST_NO_DIM.YHe / (1.0 -  CST_NO_DIM.YHe) * CST_EV_M_S_K.mass_hydrogen / CST_EV_M_S_K.mass_helium  
        Xe[:, rs[0, :] <= 3] = Xe[:, rs[0, :] <= 3] + fHe/(1+fHe)
        
    if cut_integral_min is True:
        # find the minimum and cut the integral as it is done in CLASS
        index = xp.argmin(Xe, axis=-1)
        mask = xp.arange(Xe.shape[-1]) >= index[:, None]
        Xe[mask] = 0.0
    
    # fast trapezoid integration scheme
    # h_factor_numpy is of shape (n, p), z of shape (1, p) and xHII of shape (n, p)
    # integrand is of shape (n, p), trapz of shape (n, p-1)
    # res is of shape (n,)
    integrand = Xe * (1+rs)**2 / h_factor_no_rad(rs, omega_b, omega_c, h)
    trapz     = (integrand[..., 1:] + integrand[..., :-1])/2.0
    dz        = xp.diff(rs, axis=-1)
    res       = xp.sum(trapz * dz, axis=-1)
    
    # adding the correct prefactor in front
    # pref is of shape (n,)
    pref = CST_EV_M_S_K.c_light * CST_EV_M_S_K.sigma_Thomson * n_baryons(omega_b) / (100 * h * CONVERSIONS.km_to_mpc)

    return (pref * res).flatten()



#############################################

class ShortPowerSpectrumRange(Exception):
    """
    Exception raised for too short ranges of modes to accurately compute integrals

    Attributes
    ----------
    message: str
        Explanation of the error.

    Parameters
    ----------
    scales: np.ndarray
        Range of modes we consider.
    message: str
        Explanation of the error.
    """

    def __init__(self, scales: np.ndarray, message: str ="Matter power spectrum range is too short") -> None:
        self.message = message
        self.scales  = scales
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message + "\n for scales: " + str(self.scales) + " Mpc"


## Structure formation

def check_ik_R(ik_radius: int, p: int, radius: np.ndarray) -> None:
    """
    Check that the array of mode is long enough for the radius considered.

    Parameters
    ----------
    ik_radius : int
        Index of the array of mode where equal (or closest) to the desired radius.
    p : int
        Length of the array of modes.
    radius : np.ndarray
        Array of radiuses corresponding to the modes.

    Raises
    ------
    ShortPowerSpectrumRange
        If array not long enough.
    """

    # as no way to control the precision of the integral, 
    # require that at least it is computed on enough points


    if not np.all(ik_radius > 0.1*p):
        vals_problem = radius[..., 0][ik_radius <= 0.1*p]
        if np.any(vals_problem == vals_problem): # ignore the nan
            raise ShortPowerSpectrumRange(vals_problem[vals_problem == vals_problem], "The result may be laking precision, consider running CLASS up to lower values of k (p = {})".format(p) ) 
    
    if not np.all(ik_radius < p-1):
        vals_problem = radius[..., 0][ik_radius >= p-1]
        if np.any(vals_problem == vals_problem):
            raise ShortPowerSpectrumRange(vals_problem[vals_problem == vals_problem], "The result may be laking precision, consider running CLASS up to larger values of k (p = {})".format(p)) 


def sigma_r(radius: float | np.ndarray, 
           k: np.ndarray, 
           pk: np.ndarray, 
           *, 
           window: str = 'sharpk', 
           ik_radius : np.ndarray | None = None) -> np.ndarray:
    
    """
    Standard deviation of the matter power spectrum inside `radius`.
    Note that all physical dimensions must be self consistent.
    
    Parameters
    ----------
    radius: float | numpy.ndarray 
        shape (s,) or (q1, ..., qn, r1, ..., rm, s), smoothing scale r in Mpc
    k: numpy.ndarray
        shape (q1, ..., qn, p), modes in Mpc^{-1}
    pk: numpy.ndarray
        shape (q1, ..., qn, p), power spectrum in Mpc^3
    window: str, optional
        smoothing function
    ik_radius: numpy.ndarray, optional
        shape of radius, indices of the k array corresponding to k = 1/radius
        
    Returns
    -------
    numpy.ndarray with shape (q1, ..., qn, r1, ..., rm, s) or (q1, ..., qn, s)
    """

    # make an array out of the input radius if it was not one
    radius = convert_array(radius)

    # dimension of parameters q
    n = len(k.shape) - 1               

    # if only a simple dim 1 array is passed as radius
    # add some other dummy dimensions at least of size n
    # this means however that m = 0
    if len(radius.shape) == 1 :
        radius = radius.reshape((*tuple(1 for _ in range(n)), len(radius)))

    # dimension of parameters r
    m = len(radius.shape) - len(k.shape) 
    
    # put at least shape (1, p) to k and pk
    if len(k.shape) == 1:
        k  = k[None, ...]
        pk = pk[None, ...]

    # change shape of radius to (q1, ..., qn, r1, ..., rm, s, 1)
    radius = radius[..., None]

    # get the last dimension
    p = k.shape[-1]

    # reshape k to the form (q1, ..., qn, 1, ...1, p)
    k = k.reshape(( *(k.shape[:n]), *tuple(1 for _ in range(m)), 1, p) )
    pk = pk.reshape(( *(pk.shape[:n]), *tuple(1 for _ in range(m)), 1, p) )

    # maximum bound of integration
    if window == 'sharpk':
        
        if ik_radius is None:

            ik_radius = np.argmin( (k - 1.0/radius)**2 , axis=-1)
            check_ik_R(ik_radius, p, radius)
 
    else:
        raise ValueError("no other window function that sharpk implemented yet")
    
    mask  = np.where(np.arange(p) < ik_radius[..., None], np.ones((*radius.shape[:-1], p)),  np.zeros((*radius.shape[:-1], p))) # shape (n, r, q, p)
    dlnk  = np.diff(np.log(k), axis=-1)
    
    integ = mask * (k**3) / (2*(np.pi**2)) * pk
    trapz = (integ[..., :-1] + integ[..., 1:])/2.0
    
    return np.sqrt(np.sum(trapz * dlnk, axis = -1))
        



def dsigma_r_dr(radius: float | np.ndarray, 
                k: np.ndarray, 
                pk: np.ndarray, 
                *, 
                window: str = 'sharpk', 
                sigma_radius : np.ndarray | None = None) -> np.ndarray:
    """
    Derivative of the standard deviation of the matter power spectrum 
    inside `radius`. Note that all physical dimensions must be self consistent.
    
    Parameters
    ----------
    radius: float | numpy.ndarray 
        shape (q1, ..., qn, r1, ..., rm, s), smoothing scale r in Mpc
    k: numpy.ndarray
        shape (q1, ..., qn, p), modes in Mpc^{-1}
    pk: numpy.ndarray
        shape (q1, ..., qn, p), power spectrum in Mpc^3
    window: str, optional
        smoothing function
    ik_radius: numpy.ndarray, optional
        shape of radius, indices of the k array corresponding to k = 1/radius
        
    Returns
    -------
    numpy.ndarray with shape (q1, ..., qn, r1, ..., rm, s)
    """

    m = len(radius.shape) - len(k.shape) # dimension of parameters r
    n = len(k.shape) - 1                 # dimension of parameters q
    
    # put at least shape (1, p) to k and pk
    if len(k.shape) == 1:
        k = k[None, ...]
        pk = pk[None, ...]
    
    p = k.shape[-1]

    # make an array out of the input radius if it was not one
    # change shape to (q1, ..., qn, r1, ..., rm, s, 1)
    radius = convert_array(radius)
    _radius = radius[..., None]

    # reshape k to the form (q1, ..., qn, 1, ...1, p) with m+1 1
    _k = k.reshape(( *(k.shape[:n]), *tuple(1 for _ in range(m)), 1, p) )
    
    # reshape pk to the form (q1, ..., qn, 1, ... 1, p) with m 1
    _pk = pk.reshape(( *(pk.shape[:n]), *tuple(1 for _ in range(m)), p) )

    if window == 'sharpk':
        
        # find the index of k corresponding to 1/radius
        ik_radius = np.argmin( (_k - 1.0/_radius)**2 , axis=-1)
        check_ik_R(ik_radius, p, _radius)

        if sigma_radius is None:
            sigma_radius = sigma_r(radius, k, pk, window=window, ik_radius = ik_radius)

        return  - np.take_along_axis(_pk, ik_radius, axis=-1) / radius**4 / sigma_radius / ((2*np.pi)**2)
   
    else:
        raise ValueError("no other window function that sharpk implemented yet")




def sigma_m(mass:float | np.ndarray, 
            k:np.ndarray, 
            pk:np.ndarray, 
            omega_m: float | np.ndarray, 
            *, 
            window: str = 'sharpk', 
            ik_radius: np.ndarray | None = None, 
            c: float = 2.5):
    
    """
    Standard deviation of the matter power spectrum on mass scale.
    Note that all physical dimensions must be self consistent.
    
    Parameters
    ----------
    mass: float | numpy.ndarray
        shape (q1, ..., qn, r1, ..., rm, s) mass scale in Msol
    k: numpy.ndarray
        shape (q1, ..., qn, p), modes in Mpc^{-1}
    pk: numpy.ndarray
        shape (q1, ..., qn, p), power spectrum in Mpc^3
    omega_m: float, np.ndarray 
        shape (q1, ..., qn)
    window: str, optional
        smoothing function
    ik_radius: numpy.ndarray, optional
        shape of radius, indices of the k array corresponding to k = 1/radius
    c: float, optional
        conversion factor for the mass in the sharpk window function

    Returns
    -------
    numpy.ndarray with shape (q1, ..., qn, r1, ..., rm, s)
    """

    mass    = convert_array(mass)
    omega_m = convert_array(omega_m)

    n = len(omega_m.shape)
    m = len(mass.shape) - n -1

    # put at least shape (1, p) to k and pk
    if len(k.shape) == 1:
        k = k[None, ...]
        pk = pk[None, ...]

    assert np.all(omega_m.shape == k.shape[:-1]) and  np.all(omega_m.shape == pk.shape[:-1]), "Dimension of parameters omega_m and k (or pk) should agree"

    # reshape omega_m to the of the output shape (q1, ..., qn, r1, ..., rm, s)
    omega_m = omega_m.reshape( (*(omega_m.shape), *tuple(1 for _ in range(m)), 1) )
    
    # define the averaged (background) matter density today
    rhom0 = omega_m * CST_MSOL_MPC.rho_c_over_h2

    if window == 'sharpk':
        radius = (3*mass/(4*np.pi)/rhom0)**(1/3)/c
    else:
        raise ValueError("no other window function that sharpk implemented yet")
    
    return sigma_r(radius, k, pk, window = window, ik_radius = ik_radius) 


def dsigma_m_dm(mass:float | np.ndarray, 
                k: np.ndarray, 
                pk: np.ndarray, 
                omega_m: float | np.ndarray, 
                *, 
                window: str = 'sharpk', 
                sigma_mass: np.ndarray | None = None, 
                c: float = 2.5):
    """
    Derivative of the standard deviation of the matter power spectrum 
    on mass scale. Note that all physical dimensions must be self consistent.
    
    Parameters
    ----------
    mass: float | numpy.ndarray
        shape (q1, ..., qn, r1, ..., rm, s) mass scale in Msol
    k: numpy.ndarray
        shape (q1, ..., qn, p), modes in Mpc^{-1}
    pk: numpy.ndarray
        shape (q1, ..., qn, p), power spectrum in Mpc^3
    omega_m: float, np.ndarray 
        shape (q1, ..., qn)
    window: str, optional
        smoothing function
    sigma_mass: numpy.ndarray, optional
        shape of mass, value of sigma_m at input mass
    c: float, optional
        conversion factor for the mass in the sharpk window function

    Returns
    -------
    numpy.ndarray with shape (q1, ..., qn, r1, ..., rm, s)
    """
    mass    = convert_array(mass)
    omega_m = convert_array(omega_m)

    n = len(omega_m.shape)
    m = len(mass.shape) - n -1

    # put at least shape (1, p) to k and pk
    if len(k.shape) == 1:
        k = k[None, ...]
        pk = pk[None, ...]

    assert np.all(omega_m.shape == k.shape[:-1]) and  np.all(omega_m.shape == pk.shape[:-1]), "Dimension of parameters omega_m and k (or pk) should agree"

    # reshape omega_m to the of the output shape (q1, ..., qn, r1, ..., rm, s)
    omega_m = omega_m.reshape( (*(omega_m.shape), *tuple(1 for _ in range(m)), 1) )
    
    # define the averaged (background) matter density today
    rhom0 = omega_m * CST_MSOL_MPC.rho_c_over_h2

    if window == 'sharpk':
        radius = (3*mass/(4*np.pi)/rhom0)**(1/3)/c
        drdm   = radius/(3*mass) 
    else:
        raise ValueError("no other window function that sharpk implemented yet")
    
    return dsigma_r_dr(radius, k, pk, window=window, sigma_radius=sigma_mass) * drdm


def growth_function(z: float | np.ndarray, 
                    omega_m: float | np.ndarray, 
                    h: float | np.ndarray):
    """
    Growth function of the linear density contrast

    Analatical fit from Caroll

    Parameters
    ----------
    z: float | numpy.ndarray
        shape (r,) if array, redshift range
    omega_m: float, numpy.ndarray 
        shape (q1, ..., qn), reduced matter abundance
    h: float, numpy.ndarray
        shape (q1, ..., qn), reduced hubble constant

    Returns
    -------
    numpy.ndarray with shape (q1, ..., qn, r)
    """

    z       = convert_array(z)
    h       = convert_array(h)
    omega_m = convert_array(omega_m)
    
    z = z.reshape( (*tuple(1 for _ in omega_m.shape), len(z.flatten())))

    Omega_m   = (omega_m / h**2)[..., None] 
    Omega_l   = 1.0 - Omega_m

    h_factor = (Omega_m * (1+z)**3 + Omega_l)

    Omega_m_z = Omega_m*(1+z)**3 / h_factor
    Omega_l_z = Omega_l / h_factor
    
    return 2.5*Omega_m_z/(Omega_m_z**(4.0/7.0) - Omega_l_z + (1.0 + 0.5*Omega_m_z) * (1.0 + 1.0/70.0*Omega_l_z))/(1+z)
    


def dn_dm(z: float | np.ndarray, 
          mass: float | np.ndarray, 
          k: np.ndarray, 
          pk: np.ndarray, 
          omega_m: float | np.ndarray, 
          h: float | np.ndarray, 
          sheth_a: float = 0.322,
          sheth_q: float = 1.0,
          sheth_p: float = 0.3,
          *, 
          window: str = 'sharpk', 
          c: float = 2.5):
    
    """
    Halo mass function in Msol^{-1} Mpc^{-3}

    Parameters
    ----------
    z: float | numpy.ndarray 
        shape (r,) redshift values 
    mass: float | numpy.ndarray
        shape (s,) or (q, r, s,) mass scale in Msol
    k : numpy.ndarray
        shape (q, p), modes in Mpc^{-1}
    pk : numpy.ndarray
        shape (q, p), power spectrum in Mpc^3
    omega_m: float, np.ndarray 
        shape (q,), reduced matter abundance
    h: float, numpy.ndarray
        shape (q1, ..., qn), reduced hubble constant
    window : str, optional
        smoothing function
    c : float, optional
        conversion factor for the mass in the sharpk window function

    Returns
    -------
    numpy.ndarrray with shape (q, r, s) -- in Msol / Mpc^3
    """


    mass    = convert_array(mass)    # shape (s,) or (q, r, s)
    omega_m = convert_array(omega_m) # shape (q,)
    z       = convert_array(z)       # shape (r,)

    r = len(z)
    s = len(mass)

    if len(mass.shape) == 1:
        
        q = len(omega_m)
        
        _mass = np.empty((q, r, s))
        _mass[:] = mass # shape (q, r, s)

    else:
        
        q = mass.shape[-1]
        _mass = mass


    # define the averaged (background) matter density today
    rhom0 = omega_m[:, None, None] * CST_MSOL_MPC.rho_c_over_h2 # shape (q, 1, 1)
    
    sigma    = sigma_m(_mass, k, pk, omega_m, window=window, c = c) # shape (q, r, s)
    dsigmadm = dsigma_m_dm(_mass, k, pk, omega_m, window=window, sigma_mass=sigma, c = c) # shape (q, r, s)
    growth_z = growth_function(z, omega_m, h)[:, :, None] # shape (q, r, 1)
    growth_0 = growth_function(0, omega_m, h)[:, None] # shape (q, 1, 1)
        
    one_over_sigma = np.zeros_like(sigma)
    mask = ((sigma != 0) & (sigma == sigma))
    one_over_sigma[mask]  = 1.0/sigma[mask]
    one_over_sigma[~mask] = np.nan

    nuhat = np.sqrt(sheth_q) * 1.686 * one_over_sigma * growth_0 / growth_z # shape (q, r, s)

    return -(rhom0/mass) * dsigmadm * one_over_sigma * np.sqrt(2./np.pi)* sheth_a * (1+ nuhat**(-2*sheth_p)) * nuhat * np.exp(-nuhat*nuhat/2.0)
